# The MIT License (MIT)
#
# Copyright (c) 2015 University of East Anglia, Norwich, UK
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Developed by Geoffrey French in collaboration with Dr. M. Fisher and
# Dr. M. Mackiewicz.
from typing import Optional
from PyQt5 import QtWidgets, QtWebEngineWidgets, QtGui
import click
import pathlib
from image_labelling_tool_qt import controls, web_server


class LabellerWindow (QtWidgets.QMainWindow):
    def __init__(self, tool):
        super(LabellerWindow, self).__init__(None)

        self.tool = tool

        self.setWindowTitle('Image Labeller')

        schema_json = tool.schema_store.get_schema_json()

        # Create the labeller
        self._labeller = controls.QLabellerForLabelledImages(
            server=tool.server, labelled_images=tool.labelled_images, schema=schema_json,
            tasks=tool.tasks, anno_controls=tool.anno_controls, config=tool.config, dextr_fn=tool.dextr_fn,
            enable_firebug=tool.enable_firebug)
        # Create the web engine view
        self._view = QtWebEngineWidgets.QWebEngineView()

        # If requested, use a memory-based HTTP cache
        # This is very useful if the client-side Javascript code is being developed
        # as otherwise chromium's cache will often store old versions of the code that
        # will hamper debugging and development
        if tool.use_http_memory_cache:
            self._view.page().profile().setHttpCacheType(QtWebEngineWidgets.QWebEngineProfile.MemoryHttpCache)

        # Attach the labeller to the web engine view
        self._labeller.attach_to_web_engine_view(self._view)

        # Add the QWebView to the layout
        self.setCentralWidget(self._view)

        # Switch tool action
        switch_to_schema_editor = QtWidgets.QAction(self.style().standardIcon(QtWidgets.QStyle.SP_ArrowRight), 'Switch to schema editor', self)
        switch_to_schema_editor.setStatusTip('Switch to schema editor')
        switch_to_schema_editor.triggered.connect(self.on_switch_to_schema_editor)

        menubar = self.menuBar()
        switch_tool_menu = menubar.addMenu('&Switch tool')
        switch_tool_menu.addAction(switch_to_schema_editor)

    def on_switch_to_schema_editor(self):
        self.tool.close_labeller()
        self.tool.open_schema_editor()


class SchemaEditorWindow (QtWidgets.QMainWindow):
    def __init__(self, tool):
        super(SchemaEditorWindow, self).__init__(None)

        self.tool = tool

        self.setWindowTitle('Schema editor')

        # Create the labeller
        self._schema_editor = controls.QSchemaEditorForSchemaStore(
            server=tool.server, schema_store=tool.schema_store, enable_firebug=tool.enable_firebug)
        # Create the web engine view
        self._view = QtWebEngineWidgets.QWebEngineView()

        # If requested, use a memory-based HTTP cache
        # This is very useful if the client-side Javascript code is being developed
        # as otherwise chromium's cache will often store old versions of the code that
        # will hamper debugging and development
        if tool.use_http_memory_cache:
            self._view.page().profile().setHttpCacheType(QtWebEngineWidgets.QWebEngineProfile.MemoryHttpCache)

        # Attach the labeller to the web engine view
        self._schema_editor.attach_to_web_engine_view(self._view)

        # Add the QWebView to the layout
        self.setCentralWidget(self._view)

        # Switch tool action
        switch_to_labeller = QtWidgets.QAction(self.style().standardIcon(QtWidgets.QStyle.SP_ArrowRight), 'Switch to labeller', self)
        switch_to_labeller.setStatusTip('Switch to labeller')
        switch_to_labeller.triggered.connect(self.on_switch_to_labeller)

        menubar = self.menuBar()
        switch_tool_menu = menubar.addMenu('&Switch tool')
        switch_tool_menu.addAction(switch_to_labeller)

    def on_switch_to_labeller(self):
        self.tool.close_schema_editor()
        self.tool.open_labeller()


class Tool:
    def __init__(self, server, labelled_images, schema_store, tasks, anno_controls, config,
                 dextr_fn, enable_firebug, use_http_memory_cache):
        self.server = server
        self.schema_store = schema_store
        self.labelled_images = labelled_images
        self.tasks = tasks
        self.anno_controls = anno_controls
        self.config = config
        self.dextr_fn = dextr_fn
        self.enable_firebug = enable_firebug
        self.use_http_memory_cache = use_http_memory_cache

        self._labeller = None
        self._schema_editor = None

    def open_labeller(self):
        self._labeller = LabellerWindow(self)
        self._labeller.show()

    def close_labeller(self):
        self._labeller.close()
        self._labeller = None

    def open_schema_editor(self):
        self._schema_editor = SchemaEditorWindow(self)
        self._schema_editor.show()

    def close_schema_editor(self):
        self._schema_editor.close()
        self._schema_editor = None


def _labeller_window(app, images_dir: pathlib.Path, labels_dir: Optional[pathlib.Path], schema_path: pathlib.Path,
                     readonly: bool, enable_dextr: bool, dextr_weights=Optional[pathlib.Path],
                     enable_firebug: bool = False, use_http_memory_cache: bool = False):
    from image_labelling_tool import labelled_image, labelling_tool, labelling_schema

    server = web_server.LabellerServer()
    server.start_flask_server()

    try:
        # If DEXTR is to be made available
        if enable_dextr or dextr_weights is not None:
            from dextr.model import DextrModel
            import torch

            # Load the dextr model
            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

            if dextr_weights is not None:
                dextr_weights = dextr_weights.expanduser()
                dextr_model = torch.load(dextr_weights, map_location=device)
            else:
                dextr_model = DextrModel.pascalvoc_resunet101().to(device)

            # Evaluation mode
            dextr_model.eval()

            # Define a mask prediction function
            dextr_fn = lambda image, points: dextr_model.predict([image], points[None, :, :])[0] >= 0.5
        else:
            dextr_fn = None

        schema_store = labelling_schema.FileSchemaStore(schema_path, readonly=readonly)

        # Annotation controls
        # Labels may also have optional meta-data associated with them
        # You could use this for e.g. indicating if an object is fully visible, mostly visible or significantly obscured.
        # You could also indicate quality (e.g. blurriness, etc)
        # There are four types of annotation. They have some common properties:
        #   - name: symbolic name (Python identifier)
        #   - label_text: label text in UI
        #   Check boxes, radio buttons and popup menus also have:
        #     - visibility_label_text: [optional] if provided, label visibility can be filtered by this annotation value,
        #       in which case a drop down will appear in the UI allowing the user to select a filter value
        #       that will hide/show labels accordingly
        # Control types:
        # Check box (boolean value):
        #   `labelling_tool.AnnoControlCheckbox`; only the 3 common parameters listed above
        # Radio button (choice from a list):
        #   `labelling_tool.AnnoControlRadioButtons`; the 3 common parameters listed above and:
        #       choices: list of `labelling_tool.AnnoControlRadioButtons.choice` that provide:
        #           value: symbolic value name for choice
        #           tooltip: extra information for user
        #       label_on_own_line [optional]: if True, place the label and the buttons on a separate line in the UI
        # Popup menu (choice from a grouped list):
        #   `labelling_tool.AnnoControlPopupMenu`; the 3 common parameters listed above and:
        #       groups: list of groups `labelling_tool.AnnoControlPopupMenu.group`:
        #           label_text: label text in UI
        #           choices: list of `labelling_tool.AnnoControlPopupMenu.choice` that provide:
        #               value: symbolic value name for choice
        #               label_text: choice label text in UI
        #               tooltip: extra information for user
        # Text (free form plain text):
        #   `labelling_tool.AnnoControlText`; only the 2 common parameters listed above and:
        #       - multiline: boolean; if True a text area will be used, if False a single line text entry
        anno_controls = [
            labelling_tool.AnnoControlCheckbox('good_quality', 'Good quality',
                                               visibility_label_text='Filter by good quality'),
            labelling_tool.AnnoControlRadioButtons('visibility', 'Visible', choices=[
                labelling_tool.AnnoControlRadioButtons.choice(value='full', label_text='Fully',
                                                              tooltip='Object is fully visible'),
                labelling_tool.AnnoControlRadioButtons.choice(value='mostly', label_text='Mostly',
                                                              tooltip='Object is mostly visible'),
                labelling_tool.AnnoControlRadioButtons.choice(value='obscured', label_text='Obscured',
                                                              tooltip='Object is significantly obscured'),
            ], label_on_own_line=False, visibility_label_text='Filter by visibility'),
            labelling_tool.AnnoControlPopupMenu('material', 'Material', groups=[
                labelling_tool.AnnoControlPopupMenu.group(label_text='Artifical/buildings', choices=[
                    labelling_tool.AnnoControlPopupMenu.choice(value='concrete', label_text='Concrete',
                                                               tooltip='Concrete objects'),
                    labelling_tool.AnnoControlPopupMenu.choice(value='plastic', label_text='Plastic',
                                                               tooltip='Plastic objects'),
                    labelling_tool.AnnoControlPopupMenu.choice(value='asphalt', label_text='Asphalt',
                                                               tooltip='Road, pavement, etc.'),
                ]),
                labelling_tool.AnnoControlPopupMenu.group(label_text='Flat natural', choices=[
                    labelling_tool.AnnoControlPopupMenu.choice(value='grass', label_text='Grass',
                                                               tooltip='Grass covered ground'),
                    labelling_tool.AnnoControlPopupMenu.choice(value='water', label_text='Water',
                                                               tooltip='Water/lake')]),
                labelling_tool.AnnoControlPopupMenu.group(label_text='Vegetation', choices=[
                    labelling_tool.AnnoControlPopupMenu.choice(value='trees', label_text='Trees', tooltip='Trees'),
                    labelling_tool.AnnoControlPopupMenu.choice(value='shrubbery', label_text='Shrubs',
                                                               tooltip='Shrubs/bushes'),
                    labelling_tool.AnnoControlPopupMenu.choice(value='flowers', label_text='Flowers',
                                                               tooltip='Flowers'),
                    labelling_tool.AnnoControlPopupMenu.choice(value='ivy', label_text='Ivy', tooltip='Ivy')]),
            ], visibility_label_text='Filter by material'),
            # labelling_tool.AnnoControlText('comment', 'Comment', multiline=False),
        ]

        image_paths = list(images_dir.glob('*.jpg')) + list(images_dir.glob('*.png'))
        image_paths = [p.absolute() for p in image_paths]

        # Create a `PersistentLabelledImage` instance for each image file
        labelled_images = labelled_image.LabelledImage.for_image_files(
            image_paths, labels_dir=labels_dir, readonly=bool(readonly))
        print('Loaded {0} images'.format(len(labelled_images)))

        config = web_server.DEFAULT_CONFIG

        # Example tasks to appear in checkboxes
        tasks = [
            dict(name='finished', human_name='[old] finished'),
            dict(name='segmentation', human_name='Outlines'),
            dict(name='classification', human_name='Classification'),
        ]

        tool = Tool(server=server, labelled_images=labelled_images, schema_store=schema_store,
            tasks=tasks, anno_controls=anno_controls, config=config, dextr_fn=dextr_fn,
            enable_firebug=enable_firebug, use_http_memory_cache=use_http_memory_cache)

        # Show the window and run the app
        tool.open_labeller()

        app.exec_()
    finally:
        print('Stopping server')
        server.stop_server()


@click.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
@click.option('--dextr_weights', type=click.Path())
@click.option('--enable_firebug', is_flag=True, default=False)
@click.option('--use_http_memory_cache', is_flag=True, default=False)
def run_app(dextr_weights, enable_firebug, use_http_memory_cache):
    import pathlib

    # Create an application
    app = QtWidgets.QApplication([])

    # Start with a dialog that allows the user to choose the images directory, optionally
    # a labels directory and read only checkbox
    init_dialog = QtWidgets.QDialog()
    init_dialog.setWindowTitle('Django-labeller for Qt')
    dia_layout = QtWidgets.QVBoxLayout()
    init_dialog.setLayout(dia_layout)

    # Content section
    content_section = QtWidgets.QFrame()
    dia_layout.addWidget(content_section)
    content_layout = QtWidgets.QGridLayout()
    content_section.setLayout(content_layout)

    images_dir = ['.']
    labels_dir = [None]
    schema_path = [None]

    # Images directory
    content_layout.addWidget(QtWidgets.QLabel('Please choose the directory containing the images that you wish to label:'),
                             0, 0, 1, 2)
    images_dir_button = QtWidgets.QPushButton('Choose images directory')
    images_dir_button.setIcon(init_dialog.style().standardIcon(QtWidgets.QStyle.SP_DirOpenIcon))
    images_dir_label = QtWidgets.QLabel(images_dir[0])

    def _on_images_dir_button():
        file_dialog = QtWidgets.QFileDialog(None, 'Choose images directory', '')
        file_dialog.setFileMode(QtWidgets.QFileDialog.Directory)
        choice = file_dialog.exec()
        if choice == QtWidgets.QFileDialog.Accepted:
            images_dir[0] = file_dialog.selectedFiles()[0]
            images_dir_label.setText(images_dir[0])

    images_dir_button.clicked.connect(_on_images_dir_button)
    content_layout.addWidget(images_dir_button, 1, 0)
    content_layout.addWidget(images_dir_label, 1, 1)

    # Labels directory
    content_layout.addWidget(QtWidgets.QLabel('If the labels are in a different directory please choose it:'),
                             2, 0, 1, 2)
    labels_dir_button = QtWidgets.QPushButton('[Optional] Choose labels directory')
    labels_dir_button.setIcon(init_dialog.style().standardIcon(QtWidgets.QStyle.SP_DirOpenIcon))
    labels_dir_label = QtWidgets.QLabel('')

    def _on_labels_dir_button():
        file_dialog = QtWidgets.QFileDialog(None, 'Choose labels directory', '')
        file_dialog.setFileMode(QtWidgets.QFileDialog.Directory)
        choice = file_dialog.exec()
        if choice == QtWidgets.QFileDialog.Accepted:
            labels_dir[0] = file_dialog.selectedFiles()[0]
            labels_dir_label.setText(labels_dir[0])

    labels_dir_button.clicked.connect(_on_labels_dir_button)
    content_layout.addWidget(labels_dir_button, 3, 0)
    content_layout.addWidget(labels_dir_label, 3, 1)

    # Schema path
    content_layout.addWidget(QtWidgets.QLabel('If the schema is not stored in the file schema.json alongside the labels, please choose its location:'),
                             4, 0, 1, 2)
    schema_path_button = QtWidgets.QPushButton('[Optional] Choose labels directory')
    schema_path_button.setIcon(init_dialog.style().standardIcon(QtWidgets.QStyle.SP_DirOpenIcon))
    schema_path_label = QtWidgets.QLabel('')

    def _on_schema_path_button():
        file_dialog = QtWidgets.QFileDialog(None, 'Choose schema path', '')
        # file_dialog.setFileMode(QtWidgets.QFileDialog.AnyFile)
        choice = file_dialog.exec()
        if choice == QtWidgets.QFileDialog.Accepted:
            schema_path[0] = file_dialog.selectedFiles()[0]
            schema_path_label.setText(schema_path[0])

    schema_path_button.clicked.connect(_on_schema_path_button)
    content_layout.addWidget(schema_path_button, 5, 0)
    content_layout.addWidget(schema_path_label, 5, 1)

    # Read only checkbox
    read_only_check = QtWidgets.QCheckBox('Read only')
    content_layout.addWidget(read_only_check, 6, 0)

    # DEXTR checkbox
    dextr_check = QtWidgets.QCheckBox('Enable DEXTR')
    content_layout.addWidget(dextr_check, 6, 1)


    # Buttons section
    buttons_section = QtWidgets.QFrame()
    dia_layout.addWidget(buttons_section)
    buttons_layout = QtWidgets.QHBoxLayout()
    buttons_section.setLayout(buttons_layout)

    # Buttons
    ok_button = QtWidgets.QPushButton('Ok')
    ok_button.setIcon(init_dialog.style().standardIcon(QtWidgets.QStyle.SP_DialogOkButton))

    def _on_ok():
        init_dialog.accept()

    ok_button.clicked.connect(_on_ok)
    cancel_button = QtWidgets.QPushButton('Cancel')
    cancel_button.setIcon(init_dialog.style().standardIcon(QtWidgets.QStyle.SP_DialogCancelButton))

    def _on_cancel():
        init_dialog.reject()

    cancel_button.clicked.connect(_on_cancel)

    buttons_layout.addWidget(ok_button)
    buttons_layout.addWidget(cancel_button)

    # Run the dialog
    action = init_dialog.exec()

    if action == QtWidgets.QDialog.Accepted:
        images_dir = pathlib.Path(images_dir[0])
        if labels_dir[0] is not None:
            labels_dir = pathlib.Path(labels_dir[0])
        else:
            labels_dir = None
        if schema_path[0] is not None:
            schema_path = pathlib.Path(schema_path[0])
        else:
            if labels_dir is not None:
                schema_path = labels_dir / 'schema.json'
            else:
                schema_path = images_dir / 'schema.json'
        enable_dextr = bool(dextr_check.checkState())
        readonly = bool(read_only_check.checkState())
        if dextr_weights is not None:
            dextr_weights = pathlib.Path(dextr_weights)

        _labeller_window(app, images_dir, labels_dir, schema_path, readonly, enable_dextr, dextr_weights=dextr_weights,
                         enable_firebug=enable_firebug, use_http_memory_cache=use_http_memory_cache)


if __name__ == '__main__':
    run_app()


from typing import Union
from unittest import TestCase
from . import labelling_tool


class ObjectTableTestCase(TestCase):
    class MyObject:
        def __init__(self, obj_id: Union[str, int]=None):
            self.object_id = obj_id

    def test_simple(self):
        tbl = labelling_tool.ObjectTable(None)
        a = self.MyObject()
        tbl.register(a)
        self.assertEqual(a.object_id, '{}__{}'.format(tbl._id_prefix, 1))

    def test_prefix_autoinc(self):
        tbl = labelling_tool.ObjectTable('testprefix')
        # Create objects with blank object IDs and check they are assigned IDs with auto-incrementing indices
        a = self.MyObject()
        b = self.MyObject()
        tbl.register(a)
        tbl.register(b)
        self.assertEqual(a.object_id, 'testprefix__1')
        self.assertEqual(b.object_id, 'testprefix__2')
        self.assertEqual(tbl._next_object_idx, 3)

    def test_prefix_reregister(self):
        tbl = labelling_tool.ObjectTable('testprefix')
        # Create objects with blank object IDs and check they are assigned IDs with auto-incrementing indices
        a = self.MyObject()
        b = self.MyObject()
        tbl.register(a)
        tbl.register(b)
        tbl.register(b)
        self.assertEqual(b.object_id, 'testprefix__2')

    def test_register_convert(self):
        tbl = labelling_tool.ObjectTable(None)
        # Convert old-style integer object ID to new style string '<prefix>__<index>' ID
        a = self.MyObject(12345)
        tbl.register(a)
        self.assertEqual(a.object_id, '{}__{}'.format(tbl._id_prefix, 12345))

    def test_register_duplicate(self):
        tbl = labelling_tool.ObjectTable(None)
        a = self.MyObject('abc_123')
        b = self.MyObject('abc_1234')
        c = self.MyObject('abc_1234')
        tbl.register(a)
        self.assertEqual(a.object_id, 'abc_123')
        tbl.register(b)
        self.assertEqual(b.object_id, 'abc_1234')
        self.assertRaises(ValueError, lambda: tbl.register(c))

    def test_accessors(self):
        tbl = labelling_tool.ObjectTable('xyz')
        a = self.MyObject('abc_123')
        b = self.MyObject('abc_1234')
        tbl.register(a)
        tbl.register(b)
        self.assertIs(tbl.get('abc_123'), a)
        self.assertIs(tbl.get('abc_1234'), b)
        self.assertIsNone(tbl.get(None))
        self.assertIsNone(tbl.get('xyz_789'))
        self.assertIs(tbl['abc_123'], a)
        self.assertIs(tbl['abc_1234'], b)
        self.assertIsNone(tbl[None])
        self.assertRaises(KeyError, lambda: tbl['xyz_789'])
        self.assertTrue('abc_123' in tbl)
        self.assertTrue('abc_1234' in tbl)
        self.assertFalse(None in tbl)
        self.assertFalse('xyz_789' in tbl)

    def test_new_style_id(self):
        tbl = labelling_tool.ObjectTable('pqr')
        self.assertIsNone(tbl._new_style_id(None))
        self.assertEqual(tbl._new_style_id('abc_123'), 'abc_123')
        self.assertEqual(tbl._new_style_id(123), 'pqr__123')





from typing import Any, Optional, Union, List, Tuple, Mapping, Callable
from abc import abstractmethod
import pathlib
import json


PathType = Union[pathlib.Path, str]


class ColourScheme:
    def __init__(self, name: str, human_name: str):
        self.schema = None
        self.name = name
        self.human_name = human_name

    def to_json(self) -> Any:
        return dict(name=self.name, human_name=self.human_name)

    @classmethod
    def from_json(cls, js: Any):
        return ColourScheme(js['name'], js['human_name'])


ColourTriple = Union[Tuple[int, int, int], List[int]]


class LabelClass:
    def __init__(self, name: str, human_name: str, colour: Optional[ColourTriple] = None,
                 colours: Optional[Mapping[str, ColourTriple]] = None):
        """
        Label class constructor

        Either the `colour` or `colours` parameter should be provided, not both.
        If `colour` is provided, it is converted into a dict: `{'default': colour}`

        :param name: identifier class name
        :param human_name: human readable name
        :param colour: colour as a tuple or list e.g. [255, 0, 0] for red
        :param colours: colours as a dict that maps colour scheme name to colour as a tuple or list
        """
        self.group = None
        self.name = name
        self.human_name = human_name
        if colour is not None:
            if colours is not None:
                raise TypeError('only one of colour or colours should be provided, not both')
            if isinstance(colour, (tuple, list)):
                colour = list(colour)
                if len(colour) != 3:
                    raise TypeError('colour must be a tuple or list of length 3')
                colours = {'default': colour}
            elif isinstance(colour, dict):
                colours = colour
            else:
                raise TypeError('colour should be a tuple, a list or a dict')

        if colours is not None:
            if isinstance(colours, dict):
                colours = {k: list(v) for k, v in colours.items()}
                for v in colours.values():
                    if len(v) != 3:
                        raise TypeError('values in colours must be tuples or lists of length 3')

        self.colours = colours

    def to_json(self) -> Any:
        return {'name': self.name, 'human_name': self.human_name, 'colours': self.colours}

    @classmethod
    def from_json(cls, js: Any):
        return LabelClass(js['name'], js['human_name'], colours=js['colours'])


class LabelClassGroup:
    def __init__(self, group_name: str, classes: Optional[List[LabelClass]] = None):
        """
        Label class group constructor

        :param group_name: human readable group name
        :param classes: member classes
        """
        if classes is None:
            classes = []
        self.schema = None
        self.group_name = group_name
        self.group_classes = classes
        for lcls in self.group_classes:
            lcls.group = self

    def add_class(self, lcls: LabelClass):
        lcls.group = self
        self.group_classes.append(lcls)

    def new_class(self, name: str, human_name: str, colours: Optional[Mapping[str, ColourTriple]] = None):
        lcls = LabelClass(name, human_name, colours)
        self.add_class(lcls)
        return lcls

    def to_json(self) -> Any:
        return {'group_name': self.group_name, 'group_classes': [cls.to_json() for cls in self.group_classes]}

    @classmethod
    def from_json(cls, js: Any):
        return LabelClassGroup(
            js['group_name'], [LabelClass.from_json(cls_js) for cls_js in js['group_classes']])


class LabellingSchema:
    def __init__(self, name: str, description: str, colour_schemes: Optional[List[ColourScheme]] = None,
                 label_class_groups: Optional[List[LabelClassGroup]] = None):
        if colour_schemes is None:
            colour_schemes = []
        if label_class_groups is None:
            label_class_groups = []
        self.name = name
        self.description = description
        self.colour_schemes = colour_schemes
        self.label_class_groups = label_class_groups
        for col_scheme in self.colour_schemes:
            col_scheme.schema = self
        for group in self.label_class_groups:
            group.schema = self

    def add_colour_scheme(self, col_scheme: ColourScheme):
        col_scheme.schema = self
        self.colour_schemes.append(col_scheme)

    def new_colour_scheme(self, name: str, human_name: str):
        col_scheme = ColourScheme(name, human_name)
        self.add_colour_scheme(col_scheme)
        return col_scheme

    def add_label_class_group(self, group: LabelClassGroup):
        group.schema = self
        self.label_class_groups.append(group)

    def new_label_class_group(self, group_name: str, classes: Optional[List[LabelClass]] = None):
        group = LabelClassGroup(group_name, classes)
        self.add_label_class_group(group)
        return group

    def to_json(self) -> Any:
        return {'name': self.name, 'description': self.description,
                'colour_schemes': [col_scheme.to_json() for col_scheme in self.colour_schemes],
                'label_class_groups': [cls.to_json() for cls in self.label_class_groups]}

    @classmethod
    def from_json(cls, js: Any):
        return LabellingSchema(
            js.get('name', ''), js.get('description', ''),
            [ColourScheme.from_json(col_scheme_js) for col_scheme_js in js['colour_schemes']],
            [LabelClassGroup.from_json(group_js) for group_js in js['label_class_groups']])

    @classmethod
    def empty_schema_json(cls):
        return {'name': '', 'description': '', 'colour_schemes': [], 'label_class_groups': []}

    @classmethod
    def empty(cls):
        return LabellingSchema('', '')


class SchemaStore:
    """Schema store abstract base class.

    Provides the interface by which a Flask or Qt based labelling tool accesses and updates a schema.
    """
    @abstractmethod
    def get_schema(self) -> LabellingSchema:
        """Get the schema as a `LabellingSchema` instance.
        :return: schema as a `LabellingSchema` instance.
        """
        pass

    @abstractmethod
    def get_schema_json(self) -> Any:
        """Get the schema in JSON form
        :return: schema in JSON form.
        """
        pass

    @abstractmethod
    def update_schema(self, schema: LabellingSchema):
        """Update the schema

        :param schema: updated schema
        """
        pass

    @abstractmethod
    def update_schema_json(self, schema_js: Any):
        """Update the schema in JSON form

        :param schema_js: updated schema in JSON form
        """
        pass

    @property
    def has_schema(self) -> bool:
        """Return True if schema exists

        :return: boolean
        """
        return False


class FileSchemaStore (SchemaStore):
    def __init__(self, schema_path: PathType, readonly: bool = False):
        """Labels stored in a file on disk.

        Getting the labels will load them from the file, while updating them (e.g. in response
        to user actions within the labelling tool) will write the labels to the file.

        :param schema_path: labels file path as a `str` or `pathlib.Path`
        :param readonly: if True, updating the schema via the `update_schema` or `update_schema_json` method will
            not write the schema to the file at `schema_path`
        """
        if isinstance(schema_path, str):
            schema_path = pathlib.Path(schema_path)

        self.schema_path = schema_path
        self.readonly = readonly
        self._schema = None
        self._schema_json = None

    def get_schema(self) -> LabellingSchema:
        if self._schema is None:
            if self._schema_json is None:
                if self.schema_path.exists():
                    self._schema_json = json.load(self.schema_path.open('r'))
                else:
                    self._schema_json = LabellingSchema.empty_schema_json()
            self._schema = LabellingSchema.from_json(self._schema_json)
        return self._schema

    def get_schema_json(self) -> Any:
        if self._schema_json is None:
            if self._schema is not None:
                self._schema_json = self._schema.to_json()
            else:
                if self.schema_path.exists():
                    self._schema_json = json.load(self.schema_path.open('r'))
                else:
                    self._schema_json = LabellingSchema.empty_schema_json()
        return self._schema_json

    def update_schema(self, schema: LabellingSchema):
        """Update the schema

        :param schema: updated schema
        """
        self._schema = schema
        self._schema_json = schema.to_json()
        json.dump(self._schema_json, self.schema_path.open('w'))

    def update_schema_json(self, schema_js: Any):
        """Update the schema in JSON form

        :param schema_js: updated schema in JSON form
        """
        self._schema_json = schema_js
        self._schema = None
        json.dump(self._schema_json, self.schema_path.open('w'))

    @property
    def has_schema(self) -> bool:
        """Return True if labels exist

        :return: boolean
        """
        return self.schema_path.exists()


class InMemoryLabelsStore (SchemaStore):
    def __init__(self, schema: Optional[LabellingSchema] = None,
                 on_update: Optional[Callable[[LabellingSchema], None]] = None):
        """Labels stored in memory.

        The labels are kept in memory as a `WrappedImageLabels` instance.

        :param schema: [optional] the initial schema at start
        :param on_update: [optional] a callback function of the form `fn(schema)` that will be
            invoked when the schema is updated, normally due to user actions
        """
        if schema is None:
            schema = LabellingSchema.empty()
        self.schema = schema
        self._on_update = on_update

    def get_schema(self) -> LabellingSchema:
        return self.schema

    def get_schema_json(self) -> Any:
        return self.schema.to_json()

    def update_schema(self, schema: LabellingSchema):
        """Update the schema

        :param schema: updated schema
        """
        self.schema = schema

    def update_schema_json(self, schema_js: Any):
        """Update the schema in JSON form

        :param schema_js: updated schema in JSON form
        """
        self.schema = LabellingSchema.from_json(schema_js)

    @property
    def has_schema(self) -> bool:
        """Return True if labels exist

        :return: boolean
        """
        return True



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2020-02-11 18:20
from __future__ import unicode_literals

from django.db import migrations, models, transaction


def create_default_schema(apps, schema_editor):
    LabellingSchema = apps.get_model('image_labelling_tool', 'LabellingSchema')

    with transaction.atomic():
        default_schemas = LabellingSchema.objects.filter(name='default')
        if not default_schemas.exists():
            default_schema = LabellingSchema(name='default')
            default_schema.save()


class Migration(migrations.Migration):

    dependencies = [
        ('image_labelling_tool', '0007_auto_20210322_1737'),
        ('example_labeller', '0003_dextrtask'),
    ]

    operations = [
        migrations.RunPython(create_default_schema),
    ]



# Generated by Django 3.1.7 on 2021-03-24 14:16

import django.core.validators
from django.db import migrations, models
import re


class Migration(migrations.Migration):

    dependencies = [
        ('image_labelling_tool', '0007_auto_20210322_1737'),
    ]

    operations = [
        migrations.AlterField(
            model_name='labelclass',
            name='name',
            field=models.CharField(max_length=64, validators=[django.core.validators.RegexValidator(re.compile('[a-zA-Z_][a-zA-Z0-9_]*'), message='Enter a valid identifier (letters or underscore, followedby letters, numbers or underscore')], verbose_name='Label class identifier name'),
        ),
        migrations.AlterField(
            model_name='labellingcolourscheme',
            name='name',
            field=models.CharField(max_length=64, validators=[django.core.validators.RegexValidator(re.compile('[a-zA-Z_][a-zA-Z0-9_]*'), message='Enter a valid identifier (letters or underscore, followedby letters, numbers or underscore')], verbose_name='Colour scheme identifier name'),
        ),
    ]



from __future__ import absolute_import

import os
from celery import Celery
from django.conf import settings

from kombu import serialization
serialization.registry._decoders.pop("application/x-python-serialize")

# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'example_labeller_app.settings')

app = Celery('example_labeller_app')

# Using a string here means the worker will not have to
# pickle the object when using Windows.
app.config_from_object('django.conf:settings', namespace='CELERY')
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)


def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'static',
        'dest': 'image_labelling_tool',
        'require': 'image_labelling_tool/labelling_tool/extension'
    }]


# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2017-04-11 07:40
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('image_labelling_tool', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='labels',
            name='edit_time_elapsed',
            field=models.FloatField(blank=True, default=0.0),
        ),
    ]



import os, mimetypes, json, datetime, re
import pickle
import tqdm
from django.core.management.base import BaseCommand, CommandError
from django.core.files import File
from django.db.models import Avg, Max, Min, Sum
from django.db import transaction
from image_labelling_tool import models


_DEMO_LABEL_SCHEMA = {
    'colour_schemes': [
        {'name': 'natural', 'human_name': 'Natural'},
        {'name': 'artificial', 'human_name': 'Artificial'},
    ],
    'label_class_groups': [
        {'group_name': 'Natural', 'group_classes': [
            {'name': 'tree', 'human_name': 'Trees', 'colours': dict(default=[0, 255, 192], natural=[0, 255, 192],
                                                                    artificial=[128, 128, 128])},
            {'name': 'lake', 'human_name': 'Lake', 'colours': dict(default=[0, 128, 255], natural=[0, 128, 255],
                                                                   artificial=[128, 128, 128])},
            {'name': 'flower', 'human_name': 'Flower', 'colours': dict(default=[255, 96, 192], natural=[255, 192, 96],
                                                                       artificial=[128, 128, 128])},
            {'name': 'leaf', 'human_name': 'Leaf', 'colours': dict(default=[65, 255, 0], natural=[65, 255, 0],
                                                                   artificial=[128, 128, 128])},
            {'name': 'stem', 'human_name': 'Stem', 'colours': dict(default=[128, 64, 0], natural=[128, 64, 0],
                                                                   artificial=[128, 128, 128])},
        ]},
        {'group_name': 'Artificial', 'group_classes': [
            {'name': 'building', 'human_name': 'Buildings', 'colours': dict(default=[255, 128, 0], natural=[128, 128, 128],
                                                                            artificial=[255, 128, 0])},
            {'name': 'wall', 'human_name': 'Wall', 'colours': dict(default=[0, 128, 255], natural=[128, 128, 128],
                                                                   artificial=[0, 128, 255])},
        ]},
    ]
}


class Command(BaseCommand):
    help = 'Import label classes from JSON file. Give \'default\' as the path to load the default label classes'

    def add_arguments(self, parser):
        parser.add_argument('schema_name', type=str)
        parser.add_argument('path', type=str)
        parser.add_argument('--replace', action='store_true', dest='replace', help='Delete existing label classes')

    def handle(self, *args, **options):
        schema_name = options['schema_name']
        path = options['path']
        replace = options['replace']

        with transaction.atomic():
            schema_models = models.LabellingSchema.objects.filter(name=schema_name)
            if schema_models.exists():
                schema_model = schema_models.first()
                print('Found existing schema; extending')
            else:
                print('Creating new schema named {}'.format(schema_name))
                schema_model = models.LabellingSchema(name=schema_name)
                schema_model.save()

            if replace:
                for model in models.LabelClassColour.objects.filter(label_class__group__schema=schema_model):
                    model.delete()

                for model in models.LabelClass.objects.filter(group__schema=schema_model):
                    model.delete()

                for model in models.LabelClassGroup.objects.filter(schema=schema_model):
                    model.delete()

                for model in models.LabellingColourScheme.objects.filter(schema=schema_model):
                    model.delete()

            col_scheme_order_index = models.LabellingColourScheme.objects.filter(schema=schema_model).aggregate(Max('order_index'))['order_index__max']
            if col_scheme_order_index is None:
                col_scheme_order_index = 0
            else:
                col_scheme_order_index += 1
            group_order_index = models.LabelClassGroup.objects.filter(schema=schema_model).aggregate(Max('order_index'))['order_index__max']
            if group_order_index is None:
                group_order_index = 0
            else:
                group_order_index += 1
            lab_order_index = models.LabelClass.objects.filter(group__schema=schema_model).aggregate(Max('order_index'))['order_index__max']
            if lab_order_index is None:
                lab_order_index = 0
            else:
                lab_order_index += 1

            if path == 'demo':
                schema_js = _DEMO_LABEL_SCHEMA
            else:
                with open(path, 'r') as f_in:
                    schema_js = json.load(f_in)

            groups_js = schema_js['label_class_groups']
            colour_schemes_js = schema_js['colour_schemes']

            name_to_colour_scheme = {}

            for col_scheme_js in colour_schemes_js:
                col_scheme_models = schema_model.colour_schemes.filter(name=col_scheme_js['name'])
                if col_scheme_models.exists():
                    # Update existing
                    col_scheme_model = col_scheme_models.first()
                    col_scheme_model.human_name = col_scheme_js['human_name']
                else:
                    # Create new
                    col_scheme_model = models.LabellingColourScheme(
                        schema=schema_model, name=col_scheme_js['name'], human_name=col_scheme_js['human_name'],
                        order_index=col_scheme_order_index)
                    col_scheme_order_index += 1
                col_scheme_model.save()
                name_to_colour_scheme[col_scheme_js['name']] = col_scheme_model

            for group_js in groups_js:
                group_models = schema_model.label_class_groups.filter(group_name=group_js['group_name'])
                if group_models.exists():
                    # Use existing
                    group_model = group_models.first()
                else:
                    # Create new
                    group_model = models.LabelClassGroup(
                        schema=schema_model, group_name=group_js['group_name'], order_index=group_order_index)
                    group_order_index += 1
                    group_model.save()

                for lab_js in group_js['group_classes']:
                    lab_models = models.LabelClass.objects.filter(group__schema=schema_model, name=lab_js['name'])
                    if lab_models.exists():
                        # Update existing
                        lab_model = lab_models.first()
                        lab_model.group = group_model
                        lab_model.human_name = lab_js['human_name']
                        lab_model.default_colour = models.LabelClass.list_to_html_colour(lab_js['colours']['default'])
                    else:
                        # Create new
                        lab_model = models.LabelClass(
                            group=group_model, name=lab_js['name'], human_name=lab_js['human_name'],
                            default_colour=models.LabelClass.list_to_html_colour(lab_js['colours']['default']),
                            order_index=lab_order_index)
                        lab_order_index += 1
                    lab_model.save()

                    for key, value in lab_js['colours'].items():
                        if key != 'default':
                            cls_col_models = models.LabelClassColour.objects.filter(
                                scheme=name_to_colour_scheme[key], label_class=lab_model)
                            if cls_col_models.exists():
                                # Update existing
                                cls_col_model = cls_col_models.first()
                                cls_col_model.colour = models.LabelClass.list_to_html_colour(value)
                            else:
                                # Create new
                                cls_col_model = models.LabelClassColour(
                                    label_class=lab_model, scheme=name_to_colour_scheme[key],
                                    colour=models.LabelClass.list_to_html_colour(value))
                            cls_col_model.save()




from typing import Any, TypeVar, Union, List, Dict, Optional
from abc import abstractmethod
import json
import re
from django.db import transaction
from django.http import HttpRequest, JsonResponse
from django.db.models import Avg, Max, Min, Sum
from django.views.decorators.cache import never_cache
from django.views import View
from django.utils.decorators import method_decorator
from . import models, schema_editor_messages


_INT_REGEX = re.compile(r'[\d]+')
_UUID_REGEX = re.compile(r'\b[0-9a-f]{8}\b-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-\b[0-9a-f]{12}\b')


def _update_model_from_js(model, model_attr_name, value, save=False):
    if getattr(model, model_attr_name) != value:
        setattr(model, model_attr_name, value)
        return True
    else:
        return save or False


SchemaType = TypeVar('SchemaType')


class AbstractSchemaEditorView (View, schema_editor_messages.SchemaEditorMessageHandler):
    """
    Scheme editor class based view

    Subclass and override the `get_schema` method that should retrieve the schema that we are editing.
    This will be passed to the methods that perform the udpates (`update_scheme, `create_colour_scheme`, etc.)

    Subclass and override the update methods inherited from `SchemaEditorAPI`:
    - `update_schema`
    - `create_colour_scheme`
    - `delete_colour_scheme`
    - `create_group`
    - `delete_group`
    - `create_label_class`
    - `delete_label_class`
    """
    @abstractmethod
    def get_schema(self, request: HttpRequest, *args, **kwargs) -> Union[models.LabellingSchema, SchemaType]:
        pass

    @method_decorator(never_cache)
    def post(self, request, *args, **kwargs):
        # Get the messages from the POST params
        messages = json.loads(request.POST.get('messages'))
        # Get the schema we will work with
        schema = self.get_schema(request, *args, **kwargs)
        # Handle the messages
        response = self.handle_messages(request, schema, messages)
        # Reply
        return JsonResponse(response)


class SchemaEditorView (AbstractSchemaEditorView):
    """
    Scheme editor class based view that stores schemas in the Django database using the classes
    defined in `models`.

    Subclass and override the `get_schema` method. `get_schema` should return a `models.LabellingSchema` instance.

    Example in which the URL pattern places a schema ID in the `schema_id` keyword argument:
    >>> class MySchemaEditorView (SchemaEditorView):
    ...     def get_schema(self, request: HttpRequest, *args, **kwargs) -> models.LabellingSchema:
    ...         return models.LabellingSchema.objects.get(id=kwargs['schema_id'])
    """
    @abstractmethod
    def get_schema(self, request: HttpRequest, *args, **kwargs) -> models.LabellingSchema:
        pass

    def update_schema(self, request: HttpRequest, schema: Union[models.LabellingSchema, SchemaType],
                      schema_js: Any) -> Optional[Dict[str, Dict[str, Any]]]:
        """Update all models in a given schema

        :param request: request
        :param schema: a `models.LabellingSchema` instance
        :param params: JSON representation of method parameters
        :return: JSON response
        """
        colour_scheme_id_mapping = {}
        group_id_mapping = {}
        label_class_id_mapping = {}
        with transaction.atomic():
            colour_schemes_js = schema_js.get('colour_schemes', [])
            label_class_groups_js = schema_js.get('label_class_groups', [])

            # Handle colour schemes
            for scheme_i, scheme_js in enumerate(colour_schemes_js):
                if isinstance(scheme_js['id'], str) and _UUID_REGEX.fullmatch(scheme_js['id']) is not None:
                    # New colour scheme
                    scheme_model = models.LabellingColourScheme(
                        schema=schema, name=scheme_js['name'], human_name=scheme_js['human_name'],
                        order_index=scheme_i)
                    scheme_model.save()
                    colour_scheme_id_mapping[scheme_js['id']] = scheme_model.id
                elif isinstance(scheme_js['id'], int):
                    try:
                        scheme_model = models.LabellingColourScheme.objects.get(schema=schema, id=scheme_js['id'])
                    except models.LabellingColourScheme.DoesNotExist:
                        print('WARNING: could not find colour scheme for schema={}, id={}'.format(
                            schema.name, id=scheme_js['id']
                        ))
                        scheme_model = None
                    else:
                        save = _update_model_from_js(scheme_model, 'order_index', scheme_i)
                        save = _update_model_from_js(scheme_model, 'human_name', scheme_js['human_name'], save)
                        if save:
                            scheme_model.save()
                else:
                    # No ID, ignore
                    pass

            # Handle label class groups
            for group_i, group_js in enumerate(label_class_groups_js):
                if isinstance(group_js['id'], str) and _UUID_REGEX.fullmatch(group_js['id']) is not None:
                    # New group
                    group_model = models.LabelClassGroup(
                        schema=schema, group_name=group_js['group_name'], order_index=group_i)
                    group_model.save()
                    group_id_mapping[group_js['id']] = group_model.id
                elif isinstance(group_js['id'], int):
                    try:
                        group_model = models.LabelClassGroup.objects.get(schema=schema, id=group_js['id'])
                    except models.LabelClassGroup.DoesNotExist:
                        print('WARNING: could not find group for schema={}, id={}'.format(
                            schema.name, id=group_js['id']
                        ))
                        group_model = None
                    else:
                        save = _update_model_from_js(group_model, 'order_index', group_i)
                        save = _update_model_from_js(group_model, 'group_name', group_js['group_name'], save)
                        if save:
                            group_model.save()
                else:
                    group_model = None

                if group_model is not None:
                    for lcls_i, lcls_js in enumerate(group_js['group_classes']):
                        if isinstance(lcls_js['id'], str) and _UUID_REGEX.fullmatch(lcls_js['id']) is not None:
                            # New colour scheme
                            default_col_html = models.LabelClass.list_to_html_colour(lcls_js['colours']['default'])
                            lcls_model = models.LabelClass(
                                group=group_model, name=lcls_js['name'], order_index=lcls_i,
                                human_name=lcls_js['human_name'], default_colour=default_col_html)
                            lcls_model.save()
                            label_class_id_mapping[lcls_js['id']] = lcls_model.id
                        elif isinstance(lcls_js['id'], int):
                            try:
                                lcls_model = models.LabelClass.objects.get(group__schema=schema, id=lcls_js['id'])
                            except models.LabelClass.DoesNotExist:
                                print('WARNING: could not find label class for schema={}, id={}'.format(
                                    schema.name, id=lcls_js['id']
                                ))
                                lcls_model = None
                            else:
                                default_col_html = models.LabelClass.list_to_html_colour(lcls_js['colours']['default'])
                                save = _update_model_from_js(lcls_model, 'order_index', lcls_i)
                                save = _update_model_from_js(lcls_model, 'group', group_model, save)
                                save = _update_model_from_js(lcls_model, 'human_name', lcls_js['human_name'], save)
                                save = _update_model_from_js(lcls_model, 'default_colour', default_col_html, save)
                                if save:
                                    lcls_model.save()
                        else:
                            lcls_model = None

                        if lcls_model is not None:
                            self._update_label_class_scheme_colours(schema, lcls_model, lcls_js['colours'])

        return {'colour_scheme_id_mapping': colour_scheme_id_mapping,
                'group_id_mapping': group_id_mapping,
                'label_class_id_mapping': label_class_id_mapping
                }

    def _update_label_class_scheme_colours(self, schema: models.LabellingSchema, lcls_model: models.LabelClass,
                                           colours_dict_js: Dict[str, List[int]]):
        for col_scheme_name, col_js in colours_dict_js.items():
            if col_scheme_name != 'default':
                col_models = models.LabelClassColour.objects.filter(
                    label_class=lcls_model, scheme__name=col_scheme_name)
                if col_models.exists():
                    # Existing class-colour; update
                    col_model = col_models.first()
                    col_html = models.LabelClass.list_to_html_colour(col_js)
                    save = _update_model_from_js(col_model, 'colour', col_html)
                    if save:
                        col_model.save()
                else:
                    # No existing class-colour; find the colour scheme and create new class class-colour
                    scheme_models = models.LabellingColourScheme.objects.filter(schema=schema, name=col_scheme_name)
                    if scheme_models.exists():
                        col_html = models.LabelClass.list_to_html_colour(col_js)
                        col_model = models.LabelClassColour(
                            label_class=lcls_model, scheme=scheme_models.first(), colour=col_html)
                        col_model.save()

    def create_colour_scheme(self, request: HttpRequest, schema: models.LabellingSchema,
                             colour_scheme_js: Any) -> Optional[int]:
        if schema.colour_schemes.filter(name=colour_scheme_js['name']).exists():
            # Duplicate name
            raise schema_editor_messages.NameInUseError(colour_scheme_js['name'])
        last_order_index = schema.colour_schemes.aggregate(Max('order_index'))['order_index__max']
        if last_order_index is None:
            last_order_index = 0
        colour_scheme = models.LabellingColourScheme(
            schema=schema, name=colour_scheme_js['name'], human_name=colour_scheme_js['human_name'],
            order_index=last_order_index + 1)
        colour_scheme.save()
        return colour_scheme.id

    def delete_colour_scheme(self, request: HttpRequest, schema: models.LabellingSchema, colour_scheme_js: Any):
        try:
            col_scheme = schema.colour_schemes.get(id=colour_scheme_js['id'])
        except models.LabellingColourScheme.DoesNotExist:
            raise schema_editor_messages.NotFoundError
        else:
            # Delete all class-colours that reference this colour scheme
            models.LabelClassColour.objects.filter(scheme=col_scheme).delete()
            # Now delete the colour scheme
            col_scheme.delete()

    def create_group(self, request: HttpRequest, schema: models.LabellingSchema, group_js: Any) -> Optional[int]:
        last_order_index = schema.label_class_groups.aggregate(Max('order_index'))['order_index__max']
        if last_order_index is None:
            last_order_index = 0
        group = models.LabelClassGroup(
            schema=schema, group_name=group_js['group_name'],
            order_index=last_order_index + 1)
        group.save()
        return group.id

    def delete_group(self, request: HttpRequest, schema: models.LabellingSchema,
                     group_js: Any) -> Any:
        try:
            group = schema.label_class_groups.get(id=group_js['id'])
        except models.LabelClassGroup.DoesNotExist:
            raise schema_editor_messages.NotFoundError
        else:
            if len(group.group_classes.all()) == 0:
                group.delete()
            else:
                raise schema_editor_messages.GroupNotEmptyError

    def create_label_class(self, request: HttpRequest, schema: models.LabellingSchema,
                           containing_group_js: Any, label_class_js: Any) -> Optional[int]:
        with transaction.atomic():
            if models.LabelClass.objects.filter(group__schema=schema, name=label_class_js['name']).exists():
                # Duplicate name
                raise schema_editor_messages.NameInUseError
            try:
                containing_group = schema.label_class_groups.get(id=containing_group_js['id'])
            except models.LabelClassGroup.DoesNotExist:
                raise schema_editor_messages.CouldNotFindContainingGroupError
            else:
                last_order_index = containing_group.group_classes.aggregate(Max('order_index'))['order_index__max']
                if last_order_index is None:
                    last_order_index = 0
                default_col_html = models.LabelClass.list_to_html_colour(label_class_js['colours']['default'])
                label_class = models.LabelClass(
                    group=containing_group, name=label_class_js['name'], human_name=label_class_js['human_name'],
                    default_colour=default_col_html, order_index=last_order_index + 1)
                label_class.save()
                self._update_label_class_scheme_colours(schema, label_class, label_class_js['colours'])
                return label_class.id

    def delete_label_class(self, request: HttpRequest, schema: models.LabellingSchema,
                           containing_group_js: Any, label_class_js: Any):
        try:
            label_class = models.LabelClass.objects.get(group__schema=schema, id=label_class_js['id'])
        except models.LabelClass.DoesNotExist:
            raise schema_editor_messages.NotFoundError
        else:
            # Delete all class-colours that reference this colour scheme
            models.LabelClassColour.objects.filter(label_class=label_class).delete()
            # Now delete the label class
            label_class.delete()



from celery import shared_task

import numpy as np

from image_labelling_tool import labelling_tool

from django.conf import settings


_dextr_model = None

def _apply_dextr(image_path, dextr_points_np):
    global _dextr_model
    if settings.LABELLING_TOOL_DEXTR_AVAILABLE or settings.LABELLING_TOOL_DEXTR_WEIGHTS_PATH is not None:
        if _dextr_model is None:
            import os
            from dextr.model import DextrModel
            import torch

            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

            if settings.LABELLING_TOOL_DEXTR_WEIGHTS_PATH is not None:
                dextr_weights = os.path.expanduser(settings.LABELLING_TOOL_DEXTR_WEIGHTS_PATH)
                dextr_model = torch.load(dextr_weights, map_location=device)
            else:
                dextr_model = DextrModel.pascalvoc_resunet101().to(device)

            dextr_model.eval()

            _dextr_model = dextr_model

        import numpy as np
        from PIL import Image

        im = Image.open(image_path)

        mask = _dextr_model.predict([im], dextr_points_np[None, :, :])[0] >= 0.5
        regions = labelling_tool.PolygonLabel.mask_image_to_regions_cv(mask, sort_decreasing_area=True)
        regions_js = labelling_tool.PolygonLabel.regions_to_json(regions)
        return regions_js
    else:
        return None


@shared_task
def test_task(a, b):
    return a + b

@shared_task
def dextr(image_path, dextr_points_js):
    dextr_points = np.array([[p['y'], p['x']] for p in dextr_points_js])
    regions_js = _apply_dextr(image_path, dextr_points)
    return regions_js



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2020-02-11 18:20
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('example_labeller', '0002_auto_20180911_0954'),
    ]

    operations = [
        migrations.CreateModel(
            name='DextrTask',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('creation_timestamp', models.DateTimeField(default=django.utils.timezone.now)),
                ('image_id_str', models.CharField(max_length=128)),
                ('dextr_id', models.IntegerField()),
                ('celery_task_id', models.CharField(max_length=128)),
                ('image', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='example_labeller.ImageWithLabels')),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-30 09:03
from __future__ import unicode_literals

import datetime
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Labels',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('labels_json_str', models.TextField(default=b'[]')),
                ('complete', models.BooleanField(default=False)),
                ('creation_date', models.DateField()),
                ('last_modified_datetime', models.DateTimeField(default=datetime.datetime.now)),
                ('lock_expiry_datetime', models.DateTimeField(default=datetime.datetime.now)),
                ('last_modified_by', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='modified_labels', to=settings.AUTH_USER_MODEL)),
                ('locked_by', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='locked_labels', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# Generated by Django 3.1.7 on 2021-03-22 17:37

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import re


class Migration(migrations.Migration):

    dependencies = [
        ('image_labelling_tool', '0006_remove_labels_complete'),
    ]

    operations = [
        migrations.CreateModel(
            name='LabelClass',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64, unique=True, validators=[django.core.validators.RegexValidator(re.compile('[a-zA-Z_][a-zA-Z0-9_]*'), message='Enter a valid identifier (letters or underscore, followedby letters, numbers or underscore')], verbose_name='Species identifier name')),
                ('human_name', models.CharField(default='', max_length=64)),
                ('default_colour', models.CharField(default='#0080ff', max_length=8)),
                ('order_index', models.IntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name='LabellingSchema',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(default='', max_length=256)),
                ('description', models.TextField(blank=True, default='')),
            ],
        ),
        migrations.CreateModel(
            name='LabellingColourScheme',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64, unique=True, validators=[django.core.validators.RegexValidator(re.compile('[a-zA-Z_][a-zA-Z0-9_]*'), message='Enter a valid identifier (letters or underscore, followedby letters, numbers or underscore')], verbose_name='Identifier used within tool')),
                ('human_name', models.CharField(default='', max_length=64, verbose_name='Name of colour scheme as shown in UI')),
                ('order_index', models.IntegerField(default=0)),
                ('schema', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='colour_schemes', to='image_labelling_tool.labellingschema')),
            ],
        ),
        migrations.CreateModel(
            name='LabelClassGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('group_name', models.CharField(default='', max_length=64, verbose_name='Name of label class group as shown in UI')),
                ('order_index', models.IntegerField(default=0)),
                ('schema', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='label_class_groups', to='image_labelling_tool.labellingschema')),
            ],
        ),
        migrations.CreateModel(
            name='LabelClassColour',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('colour', models.CharField(default='#0080ff', max_length=8)),
                ('label_class', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='scheme_colours', to='image_labelling_tool.labelclass')),
                ('scheme', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='label_class_colours', to='image_labelling_tool.labellingcolourscheme')),
            ],
        ),
        migrations.AddField(
            model_name='labelclass',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='group_classes', to='image_labelling_tool.labelclassgroup'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-09-11 08:54
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('image_labelling_tool', '0002_labels_edit_time_elapsed'),
    ]

    operations = [
        migrations.AlterField(
            model_name='labels',
            name='labels_json_str',
            field=models.TextField(default='[]'),
        ),
    ]



import datetime
from django.test import TestCase
from django.conf import settings
from django.contrib.auth import get_user_model
from . import models

# Create your tests here.
class LabelsMetadataTestCase(TestCase):
    def setUp(self):
        get_user_model().objects.create(username='test_user')
        models.LabellingTask.objects.create(name='test_task1', human_name='Test task 1')
        models.LabellingTask.objects.create(name='test_task2', human_name='Test task 2')

    def test_metadata_dict_to_json(self):
        test_user = get_user_model().objects.get(username='test_user')
        test_task1 = models.LabellingTask.objects.get(name='test_task1')
        test_task2 = models.LabellingTask.objects.get(name='test_task2')

        meta_js = models.Labels.metadata_dict_to_json(dict(creation_date=datetime.date(year=2020, month=2, day=21)))
        self.assertEqual(meta_js['creation_date'], '2020-02-21')

        meta_js = models.Labels.metadata_dict_to_json(dict(last_modified_by=test_user))
        self.assertEqual(meta_js['last_modified_by__username'], 'test_user')
        self.assertEqual(meta_js['last_modified_by__id'], test_user.id)

        meta_js = models.Labels.metadata_dict_to_json(dict(last_modified_datetime=datetime.datetime(
            year=2020, month=2, day=21, hour=16, minute=31, second=5)))
        self.assertEqual(meta_js['last_modified_datetime'], '2020-02-21 16:31:05')

        meta_js = models.Labels.metadata_dict_to_json(dict(completed_tasks=[]))
        self.assertEqual(meta_js['completed_tasks'], [])

        meta_js = models.Labels.metadata_dict_to_json(dict(completed_tasks=[test_task1]))
        self.assertEqual(meta_js['completed_tasks'], ['test_task1'])

        meta_js = models.Labels.metadata_dict_to_json(dict(completed_tasks=[test_task1, test_task2]))
        self.assertEqual(meta_js['completed_tasks'], ['test_task1', 'test_task2'])

    def test_metadata_json_to_dict(self):
        test_user = get_user_model().objects.get(username='test_user')
        test_task1 = models.LabellingTask.objects.get(name='test_task1')
        test_task2 = models.LabellingTask.objects.get(name='test_task2')

        metadata = models.Labels.metadata_json_to_dict(dict(creation_date='2020-02-21'))
        self.assertEqual(metadata['creation_date'], datetime.date(year=2020, month=2, day=21))

        metadata = models.Labels.metadata_json_to_dict(dict(last_modified_by__username='test_user'))
        self.assertEqual(metadata['last_modified_by'], test_user)

        metadata = models.Labels.metadata_json_to_dict(dict(last_modified_by__id=test_user.id))
        self.assertEqual(metadata['last_modified_by'], test_user)

        metadata = models.Labels.metadata_json_to_dict(dict(last_modified_datetime='2020-02-21 16:31:05'))
        self.assertEqual(metadata['last_modified_datetime'], datetime.datetime(
            year=2020, month=2, day=21, hour=16, minute=31, second=5))

        metadata = models.Labels.metadata_json_to_dict(dict(completed_tasks=[]))
        self.assertEqual(metadata['completed_tasks'], [])

        metadata = models.Labels.metadata_json_to_dict(dict(completed_tasks=['test_task1']))
        self.assertEqual(metadata['completed_tasks'], [test_task1])

        metadata = models.Labels.metadata_json_to_dict(dict(completed_tasks=['test_task1', 'test_task2']))
        self.assertEqual(metadata['completed_tasks'], [test_task1, test_task2])




import math
import numpy as np
from PIL import Image, ImageDraw
from unittest import TestCase
from . import labelling_tool


class AbstractLabelTestCase(TestCase):
    def test_constructor(self):
        a = labelling_tool.AbstractLabel()
        b = labelling_tool.AbstractLabel(object_id='abc_123', classification='cls_a',
                                         source='manual', anno_data={'purpose': 'test'})
        self.assertIsNone(a.object_id)
        self.assertEqual(b.object_id, 'abc_123')
        self.assertIsNone(a.classification)
        self.assertEqual(b.classification, 'cls_a')
        self.assertIsNone(a.source)
        self.assertEqual(b.source, 'manual')
        self.assertEqual(a.anno_data, {})
        self.assertEqual(b.anno_data, {'purpose': 'test'})

    def test_dependencies(self):
        a = labelling_tool.AbstractLabel()
        self.assertEqual(a.dependencies, [])

    def test_flatten(self):
        a = labelling_tool.AbstractLabel()
        self.assertEqual(list(a.flatten()), [a])

    def test_accumulate_label_class_histogram(self):
        a = labelling_tool.AbstractLabel()
        b = labelling_tool.AbstractLabel(object_id='abc_123', classification='cls_a',
                                         source='manual', anno_data={'purpose': 'test'})
        h1 = {}
        a.accumulate_label_class_histogram(h1)
        self.assertEqual(h1, {None: 1})
        h2 = {}
        b.accumulate_label_class_histogram(h2)
        self.assertEqual(h2, {'cls_a': 1})
        b.accumulate_label_class_histogram(h2)
        self.assertEqual(h2, {'cls_a': 2})


class PointLabelTestCase(TestCase):
    def test_constructor(self):
        a = labelling_tool.PointLabel(position_xy=np.array([-1.0, 1.0]))
        b = labelling_tool.PointLabel(position_xy=np.array([1.0, 2.0]), object_id='abc_123', classification='cls_a',
                                      source='manual', anno_data={'purpose': 'test'})
        c = labelling_tool.PointLabel(position_xy=np.array([2.0, 4.0]), classification='cls_b',
                                      source='auto:test', anno_data={'purpose': 'second_test'})
        self.assertTrue((a.position_xy == np.array([-1.0, 1.0])).all())
        self.assertTrue((b.position_xy == np.array([1.0, 2.0])).all())
        self.assertTrue((c.position_xy == np.array([2.0, 4.0])).all())
        self.assertIsNone(a.object_id)
        self.assertEqual(b.object_id, 'abc_123')
        self.assertIsNone(c.object_id)
        self.assertIsNone(a.classification)
        self.assertEqual(b.classification, 'cls_a')
        self.assertEqual(c.classification, 'cls_b')
        self.assertIsNone(a.source)
        self.assertEqual(b.source, 'manual')
        self.assertEqual(c.source, 'auto:test')
        self.assertEqual(a.anno_data, {})
        self.assertEqual(b.anno_data, {'purpose': 'test'})
        self.assertEqual(c.anno_data, {'purpose': 'second_test'})

    def test_bounding_box(self):
        rad01 = labelling_tool.LabelContext(point_radius=0.1)
        a = labelling_tool.PointLabel(position_xy=np.array([-1.0, 1.0]))
        self.assertTrue((a.bounding_box()[0] == np.array([-1.0, 1.0])).all())
        self.assertTrue((a.bounding_box()[1] == np.array([-1.0, 1.0])).all())
        self.assertTrue((a.bounding_box(rad01)[0] == np.array([-1.1, 0.9])).all())
        self.assertTrue((a.bounding_box(rad01)[1] == np.array([-0.9, 1.1])).all())

    def test_warped(self):
        obj_tab = labelling_tool.ObjectTable('abc')
        a = labelling_tool.PointLabel(position_xy=np.array([-1.0, 1.0]))
        a_7 = a.warped(lambda p_xy: p_xy + 7.0)
        a_7b = a.warped(lambda p_xy: p_xy + 7.0, object_table=obj_tab)
        a_7c = a.warped(lambda p_xy: p_xy + 7.0, id_prefix='abc')
        self.assertTrue((a.position_xy == np.array([-1.0, 1.0])).all())
        self.assertTrue((a_7.position_xy == np.array([6.0, 8.0])).all())
        self.assertTrue((a_7b.position_xy == np.array([6.0, 8.0])).all())
        self.assertIsNone(a.object_id)
        self.assertTrue(a_7.object_id.endswith('__1'))
        self.assertEqual(a_7b.object_id, 'abc__1')
        self.assertEqual(a_7c.object_id, 'abc__1')

    def test_render_mask(self):
        rad10 = labelling_tool.LabelContext(point_radius=10.0)
        a = labelling_tool.PointLabel(position_xy=np.array([25.0, 25.0]))
        # Point
        tgt_point = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_point).point((25, 25), fill=1)
        self.assertTrue((a.render_mask(50, 50, fill=False, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_point)).all())
        self.assertTrue((a.render_mask(50, 50, fill=True, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_point)).all())
        # Radius 10 outlined
        tgt_rad10_outline = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_rad10_outline).ellipse([15, 15, 35, 35], outline=1, fill=0)
        self.assertTrue((a.render_mask(50, 50, fill=False, dx=0.0, dy=0.0, ctx=rad10) ==
                         np.array(tgt_rad10_outline)).all())
        # Radius 10 filled
        tgt_rad10_filled = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_rad10_filled).ellipse([15, 15, 35, 35], outline=1, fill=1)
        self.assertTrue((a.render_mask(50, 50, fill=True, dx=0.0, dy=0.0, ctx=rad10) ==
                         np.array(tgt_rad10_filled)).all())
        # Point, offset
        tgt_point_dxy = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_point_dxy).point((20, 30), fill=1)
        self.assertTrue((a.render_mask(50, 50, fill=False, dx=-5.0, dy=5.0, ctx=None) ==
                         np.array(tgt_point_dxy)).all())
        # Radius 10 filled, offset
        tgt_rad10_filled_dxy = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_rad10_filled_dxy).ellipse([10, 20, 30, 40], outline=1, fill=1)
        self.assertTrue((a.render_mask(50, 50, fill=True, dx=-5.0, dy=5.0, ctx=rad10) ==
                         np.array(tgt_rad10_filled_dxy)).all())

    def test_to_json(self):
        b = labelling_tool.PointLabel(position_xy=np.array([1.0, 2.0]), object_id='abc_123', classification='cls_a',
                                      source='manual', anno_data={'purpose': 'test'})
        self.assertEqual(b.to_json(),
                         dict(label_type='point', position={'x': 1.0, 'y': 2.0}, object_id='abc_123',
                              label_class='cls_a', source='manual', anno_data={'purpose': 'test'}))

    def test_from_json(self):
        obj_tab = labelling_tool.ObjectTable('abc')
        js = dict(label_type='point', position={'x': 1.0, 'y': 2.0}, object_id='abc_123',
                  label_class='cls_a', source='manual', anno_data={'purpose': 'test'})
        b = labelling_tool.AbstractLabel.from_json(js, obj_tab)
        self.assertTrue(isinstance(b, labelling_tool.PointLabel))
        self.assertTrue((b.position_xy == np.array([1.0, 2.0])).all())
        self.assertEqual(b.object_id, 'abc_123')
        self.assertEqual(b.classification, 'cls_a')
        self.assertEqual(b.source, 'manual')
        self.assertEqual(b.anno_data, {'purpose': 'test'})

    def test_flatten_json(self):
        js = dict(label_type='point', position={'x': 1.0, 'y': 2.0}, object_id='abc_123',
                  label_class='cls_a', source='manual', anno_data={'purpose': 'test'})
        labs_js = list(labelling_tool.AbstractLabel.flatten_json(js))
        self.assertEqual(labs_js, [js])


class PolygonLabelTestCase(TestCase):
    def are_polygons_cyclically_equal(self, a, b, both_directions=False):
        if len(a) == len(b):
            for i in range(len(a)):
                if (a == np.append(b[i:], b[:i], axis=0)).all():
                    return True
            if both_directions:
                a = a[::-1]
                for i in range(len(a)):
                    if (a == np.append(b[i:], b[:i], axis=0)).all():
                        return True
        return False

    def test_constructor(self):
        outer_rect = np.array([[10.0, 10.0], [40.0, 10.0], [40.0, 40.0], [10.0, 40.0]])
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        a = labelling_tool.PolygonLabel(regions=[inner_rect], object_id='abc_123', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        b = labelling_tool.PolygonLabel(regions=[outer_rect, inner_rect], object_id='abc_123', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        self.assertEqual(len(a.regions), 1)
        self.assertTrue((a.regions[0] == inner_rect).all())
        self.assertEqual(len(b.regions), 2)
        self.assertTrue((b.regions[0] == outer_rect).all())
        self.assertTrue((b.regions[1] == inner_rect).all())
        self.assertEqual(b.object_id, 'abc_123')
        self.assertEqual(b.classification, 'cls_a')
        self.assertEqual(b.source, 'manual')
        self.assertEqual(b.anno_data, {'purpose': 'test'})

    def test_bounding_box(self):
        outer_rect = np.array([[10.0, 10.0], [40.0, 10.0], [40.0, 40.0], [10.0, 40.0]])
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        a = labelling_tool.PolygonLabel(regions=[inner_rect], object_id='abc_123', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        b = labelling_tool.PolygonLabel(regions=[outer_rect, inner_rect], object_id='abc_123', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        self.assertTrue((a.bounding_box()[0] == np.array([20.0, 20.0])).all())
        self.assertTrue((a.bounding_box()[1] == np.array([30.0, 30.0])).all())
        self.assertTrue((b.bounding_box()[0] == np.array([10.0, 10.0])).all())
        self.assertTrue((b.bounding_box()[1] == np.array([40.0, 40.0])).all())

    def test_warped(self):
        outer_rect = np.array([[10.0, 10.0], [40.0, 10.0], [40.0, 40.0], [10.0, 40.0]])
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        a = labelling_tool.PolygonLabel(regions=[inner_rect], object_id='abc_123', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        b = labelling_tool.PolygonLabel(regions=[outer_rect, inner_rect], object_id='abc_123', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        a_7 = a.warped(lambda p_xy: p_xy + 7.0)
        b_7 = b.warped(lambda p_xy: p_xy + 7.0)
        self.assertEqual(len(a_7.regions), 1)
        self.assertTrue((a_7.regions[0] == (inner_rect + 7)).all())
        self.assertEqual(len(b_7.regions), 2)
        self.assertTrue((b_7.regions[0] == (outer_rect + 7)).all())
        self.assertTrue((b_7.regions[1] == (inner_rect + 7)).all())

        # Rotation matrix
        theta = np.radians(20.0)
        c = np.cos(theta)
        s = np.sin(theta)
        r = np.array([[c, -s],
                      [s, c]])
        a_r = a.warped(lambda p_xy: (r @ p_xy.T).T)
        self.assertTrue(np.allclose(a_r.regions[0], (r @ inner_rect.T).T))

    def test_render_mask(self):
        outer_rect = np.array([[10.0, 10.0], [40.0, 10.0], [40.0, 40.0], [10.0, 40.0]])
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        a = labelling_tool.PolygonLabel(regions=[inner_rect], object_id='abc_123', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        b = labelling_tool.PolygonLabel(regions=[outer_rect, inner_rect], object_id='abc_123', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        # Outlined
        tgt_a_outline = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_a_outline).polygon([tuple(v) for v in inner_rect], outline=1, fill=0)
        self.assertTrue((a.render_mask(50, 50, fill=False, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_a_outline)).all())

        tgt_b_outline = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_b_outline).polygon([tuple(v) for v in outer_rect], outline=1, fill=0)
        ImageDraw.Draw(tgt_b_outline).polygon([tuple(v) for v in inner_rect], outline=1, fill=0)
        self.assertTrue((b.render_mask(50, 50, fill=False, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_b_outline)).all())

        # Outlined, offset
        tgt_b_outline_dxy = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_b_outline_dxy).polygon([tuple(v) for v in outer_rect + np.array([5, -5])], outline=1, fill=0)
        ImageDraw.Draw(tgt_b_outline_dxy).polygon([tuple(v) for v in inner_rect + np.array([5, -5])], outline=1, fill=0)
        self.assertTrue((b.render_mask(50, 50, fill=False, dx=5.0, dy=-5.0, ctx=None) ==
                         np.array(tgt_b_outline_dxy)).all())

        # Filled
        tgt_a_filled = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_a_filled).polygon([tuple(v) for v in inner_rect], outline=1, fill=1)
        self.assertTrue((a.render_mask(50, 50, fill=True, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_a_filled)).all())

        tgt_b_filled = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_b_filled).polygon([tuple(v) for v in outer_rect], outline=1, fill=1)
        ImageDraw.Draw(tgt_b_filled).polygon([tuple(v) for v in inner_rect], outline=0, fill=0)
        self.assertTrue((b.render_mask(50, 50, fill=True, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_b_filled)).all())

        # Filled, offset
        tgt_b_filled_dxy = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_b_filled_dxy).polygon([tuple(v) for v in outer_rect + np.array([5, -5])], outline=1, fill=1)
        ImageDraw.Draw(tgt_b_filled_dxy).polygon([tuple(v) for v in inner_rect + np.array([5, -5])], outline=0, fill=0)
        self.assertTrue((b.render_mask(50, 50, fill=True, dx=5.0, dy=-5.0, ctx=None) ==
                         np.array(tgt_b_filled_dxy)).all())

    def test_to_json(self):
        outer_rect = np.array([[10.0, 10.0], [40.0, 10.0], [40.0, 40.0], [10.0, 40.0]])
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        a = labelling_tool.PolygonLabel(regions=[inner_rect], object_id='abc_123', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        b = labelling_tool.PolygonLabel(regions=[outer_rect, inner_rect], object_id='abc_123', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})

        outer_js = [dict(x=p[0], y=p[1]) for p in outer_rect]
        inner_js = [dict(x=p[0], y=p[1]) for p in inner_rect]
        self.assertEqual(a.to_json(),
                         dict(label_type='polygon', regions=[inner_js], object_id='abc_123',
                              label_class='cls_a', source='manual', anno_data={'purpose': 'test'}))
        self.assertEqual(b.to_json(),
                         dict(label_type='polygon', regions=[outer_js, inner_js], object_id='abc_123',
                              label_class='cls_a', source='manual', anno_data={'purpose': 'test'}))

    def test_from_json(self):
        outer_rect = np.array([[10.0, 10.0], [40.0, 10.0], [40.0, 40.0], [10.0, 40.0]])
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])

        outer_js = [dict(x=p[0], y=p[1]) for p in outer_rect]
        inner_js = [dict(x=p[0], y=p[1]) for p in inner_rect]

        obj_tab = labelling_tool.ObjectTable('abc')
        js_a = dict(label_type='polygon', regions=[inner_js], object_id='abc_123',
                    label_class='cls_a', source='manual', anno_data={'purpose': 'test'})
        js_b = dict(label_type='polygon', regions=[outer_js, inner_js], object_id='abc_124',
                    label_class='cls_b', source='manual2', anno_data={'purpose': 'test2'})
        a = labelling_tool.AbstractLabel.from_json(js_a, obj_tab)
        b = labelling_tool.AbstractLabel.from_json(js_b, obj_tab)
        self.assertTrue(isinstance(a, labelling_tool.PolygonLabel))
        self.assertTrue(isinstance(b, labelling_tool.PolygonLabel))
        self.assertEqual(len(a.regions), 1)
        self.assertTrue((a.regions[0] == inner_rect).all())
        self.assertEqual(len(b.regions), 2)
        self.assertTrue((b.regions[0] == outer_rect).all())
        self.assertTrue((b.regions[1] == inner_rect).all())
        self.assertEqual(a.object_id, 'abc_123')
        self.assertEqual(b.object_id, 'abc_124')
        self.assertEqual(a.classification, 'cls_a')
        self.assertEqual(b.classification, 'cls_b')
        self.assertEqual(a.source, 'manual')
        self.assertEqual(b.source, 'manual2')
        self.assertEqual(a.anno_data, {'purpose': 'test'})
        self.assertEqual(b.anno_data, {'purpose': 'test2'})

    def test_mask_image_to_regions(self):
        outer_rect = np.array([[10.0, 10.0], [40.0, 10.0], [40.0, 40.0], [10.0, 40.0]])
        # skimage.measure.find_contours rounds corners
        outer_rect_v2 = np.array([[10.0, 9.5], [40.0, 9.5], [40.5, 10.0], [40.5, 40.0],
                                  [40.0, 40.5], [10.0, 40.5], [9.5, 40.0], [9.5, 10.0]])
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        inner_rect_v2 = np.array([[20.0, 19.5], [30.0, 19.5], [30.5, 20.0], [30.5, 30.0],
                                  [30.0, 30.5], [20.0, 30.5], [19.5, 30.0], [19.5, 20.0]])

        tgt_a_filled = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_a_filled).polygon([tuple(v) for v in inner_rect], outline=1, fill=1)
        reg_a = labelling_tool.PolygonLabel.mask_image_to_regions(np.array(tgt_a_filled) != 0)
        self.assertEqual(len(reg_a), 1)
        self.assertTrue(self.are_polygons_cyclically_equal(reg_a[0], inner_rect_v2, both_directions=True))

        tgt_b_filled = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_b_filled).polygon([tuple(v) for v in outer_rect], outline=1, fill=1)
        ImageDraw.Draw(tgt_b_filled).polygon([tuple(v) for v in inner_rect], outline=0, fill=0)
        reg_b = labelling_tool.PolygonLabel.mask_image_to_regions(np.array(tgt_b_filled) != 0)
        self.assertEqual(len(reg_b), 2)
        self.assertTrue(self.are_polygons_cyclically_equal(reg_b[0], outer_rect_v2, both_directions=True))
        self.assertTrue(self.are_polygons_cyclically_equal(reg_b[1], inner_rect_v2, both_directions=True))

    def test_mask_image_to_regions_cv(self):
        outer_rect = np.array([[10.0, 10.0], [40.0, 10.0], [40.0, 40.0], [10.0, 40.0]])
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        # cv2.findContours generates a slightly rotated square for the inner region...
        inner_rect_v2 = np.array([[20.0, 19.0], [31.0, 20.0], [30.0, 31.0], [19.0, 30.0]])

        tgt_a_filled = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_a_filled).polygon([tuple(v) for v in inner_rect], outline=1, fill=1)
        reg_a = labelling_tool.PolygonLabel.mask_image_to_regions_cv(np.array(tgt_a_filled) != 0)
        self.assertEqual(len(reg_a), 1)
        self.assertTrue(self.are_polygons_cyclically_equal(reg_a[0], inner_rect.astype(int), both_directions=True))

        tgt_b_filled = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_b_filled).polygon([tuple(v) for v in outer_rect], outline=1, fill=1)
        ImageDraw.Draw(tgt_b_filled).polygon([tuple(v) for v in inner_rect], outline=0, fill=0)
        reg_b = labelling_tool.PolygonLabel.mask_image_to_regions_cv(np.array(tgt_b_filled) != 0)
        self.assertEqual(len(reg_b), 2)
        self.assertTrue(self.are_polygons_cyclically_equal(reg_b[0], outer_rect.astype(int), both_directions=True))
        self.assertTrue(self.are_polygons_cyclically_equal(reg_b[1], inner_rect_v2.astype(int), both_directions=True))


class BoxLabelTestCase(TestCase):
    def test_constructor(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]),
                                    object_id='abc_123', classification='cls_a',
                                    source='manual', anno_data={'purpose': 'test'})
        self.assertTrue((a.centre_xy == np.array([15.0, 25.0])).all())
        self.assertTrue((a.size_xy == np.array([8.0, 12.0])).all())
        self.assertEqual(a.object_id, 'abc_123')
        self.assertEqual(a.classification, 'cls_a')
        self.assertEqual(a.source, 'manual')
        self.assertEqual(a.anno_data, {'purpose': 'test'})

    def test_bounding_box(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]))
        self.assertTrue((a.bounding_box()[0] == np.array([11.0, 19.0])).all())
        self.assertTrue((a.bounding_box()[1] == np.array([19.0, 31.0])).all())

    def test_warped(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]))

        a_7 = a.warped(lambda p_xy: p_xy + 7.0)
        self.assertTrue((a_7.centre_xy == np.array([22.0, 32.0])).all())
        self.assertTrue((a_7.size_xy == np.array([8.0, 12.0])).all())

        # Rotation matrix
        theta = np.radians(20.0)
        c = np.cos(theta)
        s = np.sin(theta)
        r = np.array([[c, -s],
                      [s, c]])
        a_r = a.warped(lambda p_xy: (r @ p_xy.T).T)
        self.assertTrue(np.allclose(a_r.centre_xy, r @ np.array([15.0, 25.0])))
        self.assertTrue(np.allclose(a_r.size_xy, np.array([8.0 * c + 12.0 * s, 12.0 * c + 8.0 * s])))

    def test_render_mask(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]))

        # Outlined
        tgt_a_outline = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_a_outline).rectangle([(11.0, 19.0), (19.0, 31.0)], outline=1, fill=0)
        self.assertTrue((a.render_mask(50, 50, fill=False, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_a_outline)).all())

        # Outlined, offset
        tgt_b_outline_dxy = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_b_outline_dxy).rectangle([(16.0, 14.0), (24.0, 26.0)], outline=1, fill=0)
        self.assertTrue((a.render_mask(50, 50, fill=False, dx=5.0, dy=-5.0, ctx=None) ==
                         np.array(tgt_b_outline_dxy)).all())

        # Filled
        tgt_a_filled = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_a_filled).rectangle([(11.0, 19.0), (19.0, 31.0)], outline=1, fill=1)
        self.assertTrue((a.render_mask(50, 50, fill=True, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_a_filled)).all())

        # Filled, offset
        tgt_b_filled_dxy = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_b_filled_dxy).rectangle([(16.0, 14.0), (24.0, 26.0)], outline=1, fill=1)
        self.assertTrue((a.render_mask(50, 50, fill=True, dx=5.0, dy=-5.0, ctx=None) ==
                         np.array(tgt_b_filled_dxy)).all())

    def test_to_json(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]),
                                    object_id='abc_123', classification='cls_a',
                                    source='manual', anno_data={'purpose': 'test'})

        self.assertEqual(a.to_json(),
                         dict(label_type='box', centre=dict(x=15.0, y=25.0), size=dict(x=8.0, y=12.0),
                              object_id='abc_123', label_class='cls_a', source='manual',
                              anno_data={'purpose': 'test'}))

    def test_from_json(self):
        obj_tab = labelling_tool.ObjectTable('abc')
        js_a = dict(label_type='box', centre=dict(x=15.0, y=25.0), size=dict(x=8.0, y=12.0),
                    object_id='abc_123', label_class='cls_a', source='manual', anno_data={'purpose': 'test'})
        a = labelling_tool.AbstractLabel.from_json(js_a, obj_tab)
        self.assertTrue(isinstance(a, labelling_tool.BoxLabel))
        self.assertTrue((a.centre_xy == np.array([15.0, 25.0])).all())
        self.assertTrue((a.size_xy == np.array([8.0, 12.0])).all())
        self.assertEqual(a.object_id, 'abc_123')
        self.assertEqual(a.classification, 'cls_a')
        self.assertEqual(a.source, 'manual')
        self.assertEqual(a.anno_data, {'purpose': 'test'})


class OrientedEllipseLabelTestCase(TestCase):
    def test_constructor(self):
        a = labelling_tool.OrientedEllipseLabel(
            centre_xy=np.array([15.0, 25.0]), radius1=10.0, radius2=3.0, orientation_rad=math.radians(30.0),
            object_id='abc_123', classification='cls_a', source='manual', anno_data={'purpose': 'test'})
        self.assertTrue((a.centre_xy == np.array([15.0, 25.0])).all())
        self.assertEqual(a.radius1, 10.0)
        self.assertEqual(a.radius2, 3.0)
        self.assertEqual(a.orientation_rad, math.radians(30.0))
        self.assertEqual(a.object_id, 'abc_123')
        self.assertEqual(a.classification, 'cls_a')
        self.assertEqual(a.source, 'manual')
        self.assertEqual(a.anno_data, {'purpose': 'test'})

    def test_bounding_box(self):
        # Ellipse with no orientation
        a = labelling_tool.OrientedEllipseLabel(
            centre_xy=np.array([15.0, 25.0]), radius1=4.0, radius2=6.0, orientation_rad=0.0)
        self.assertTrue(np.allclose(a.bounding_box()[0], np.array([11.0, 19.0])))
        self.assertTrue(np.allclose(a.bounding_box()[1], np.array([19.0, 31.0])))
        # 90 degrees orientation
        b = labelling_tool.OrientedEllipseLabel(
            centre_xy=np.array([15.0, 25.0]), radius1=4.0, radius2=6.0, orientation_rad=math.radians(90.0))
        self.assertTrue(np.allclose(b.bounding_box()[0], np.array([9.0, 21.0])))
        self.assertTrue(np.allclose(b.bounding_box()[1], np.array([21.0, 29.0])))
        # 45 degrees orientation
        c = labelling_tool.OrientedEllipseLabel(
            centre_xy=np.array([15.0, 25.0]), radius1=4.0, radius2=6.0, orientation_rad=math.radians(45.0))
        self.assertTrue(np.allclose(c.bounding_box()[0], np.array([9.900980486407214, 19.900980486407214])))
        self.assertTrue(np.allclose(c.bounding_box()[1], np.array([20.099019513592786, 30.099019513592786])))
        # 30 degrees orientation
        d = labelling_tool.OrientedEllipseLabel(
            centre_xy=np.array([15.0, 25.0]), radius1=4.0, radius2=6.0, orientation_rad=math.radians(30.0))
        self.assertTrue(np.allclose(d.bounding_box()[0], np.array([10.41742430504416, 19.432235637169978])))
        self.assertTrue(np.allclose(d.bounding_box()[1], np.array([19.58257569495584, 30.567764362830022])))

    def test_warped(self):
        a = labelling_tool.OrientedEllipseLabel(
            centre_xy=np.array([15.0, 25.0]), radius1=4.0, radius2=6.0, orientation_rad=0.0)

        a_7 = a.warped(lambda p_xy: p_xy + 7.0)
        self.assertTrue(np.allclose(a_7.centre_xy, np.array([22.0, 32.0])))
        self.assertTrue(np.allclose(a_7.radius1, 4.0 ))
        self.assertTrue(np.allclose(a_7.radius2, 6.0 ))
        self.assertTrue(np.allclose(a_7.orientation_rad, 0.0))

        # Rotation matrix
        theta = np.radians(20.0)
        c = np.cos(theta)
        s = np.sin(theta)
        r = np.array([[c, -s],
                      [s, c]])
        a_r = a.warped(lambda p_xy: (r @ p_xy.T).T)
        self.assertTrue(np.allclose(a_r.centre_xy, r @ np.array([15.0, 25.0])))
        self.assertTrue(np.allclose(a_r.radius1, 4.0))
        self.assertTrue(np.allclose(a_r.radius2, 6.0))
        self.assertTrue(np.allclose(a_r.orientation_rad, math.radians(20.0)))

    def test_render_mask(self):
        def draw_polygon_image(verts_xy, outline, fill, image_size):
            img = Image.new('L', image_size, 0)
            xy = [(v[0], v[1]) for v in verts_xy]
            ImageDraw.Draw(img).polygon(xy, outline=outline, fill=fill)
            return img

        a = labelling_tool.OrientedEllipseLabel(
            centre_xy=np.array([15.0, 25.0]), radius1=4.0, radius2=6.0, orientation_rad=0.0)

        # 38 vertices for
        n_verts = 38
        thetas = np.linspace(0.0, math.pi * 2, n_verts + 1)[:-1]
        # Axis aligned verts, centred on origin
        aa_verts_xy = np.stack([np.cos(thetas) * 4.0, np.sin(thetas) * 6.0], axis=1)

        # Outlined
        tgt_a_outline = draw_polygon_image(aa_verts_xy + np.array([15.0, 25.0]),
                                           outline=1, fill=0, image_size=(50,50))
        self.assertTrue((a.render_mask(50, 50, fill=False, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_a_outline)).all())

        # Outlined, offset
        tgt_a_outline_dxy = draw_polygon_image(aa_verts_xy + np.array([20.0, 20.0]),
                                               outline=1, fill=0, image_size=(50,50))
        self.assertTrue((a.render_mask(50, 50, fill=False, dx=5.0, dy=-5.0, ctx=None) ==
                         np.array(tgt_a_outline_dxy)).all())

        # Filled
        tgt_a_filled = draw_polygon_image(aa_verts_xy + np.array([15.0, 25.0]),
                                          outline=1, fill=1, image_size=(50,50))
        self.assertTrue((a.render_mask(50, 50, fill=True, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_a_filled)).all())

        # Filled, offset
        tgt_b_filled_dxy = draw_polygon_image(aa_verts_xy + np.array([20.0, 20.0]),
                                              outline=1, fill=1, image_size=(50,50))
        self.assertTrue((a.render_mask(50, 50, fill=True, dx=5.0, dy=-5.0, ctx=None) ==
                         np.array(tgt_b_filled_dxy)).all())

        # With orientation
        b = labelling_tool.OrientedEllipseLabel(
            centre_xy=np.array([15.0, 25.0]), radius1=4.0, radius2=6.0, orientation_rad=math.radians(25.0))

        rot_theta = np.radians(25.0)
        c = np.cos(rot_theta)
        s = np.sin(rot_theta)
        r = np.array([[c, -s],
                      [s, c]])

        # Outlined
        tgt_b_outline = draw_polygon_image((r @ aa_verts_xy.T).T + np.array([15.0, 25.0]),
                                           outline=1, fill=0, image_size=(50,50))
        self.assertTrue((b.render_mask(50, 50, fill=False, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_b_outline)).all())

        # Outlined, offset
        tgt_b_outline_dxy = draw_polygon_image((r @ aa_verts_xy.T).T + np.array([20.0, 20.0]),
                                               outline=1, fill=0, image_size=(50,50))
        self.assertTrue((b.render_mask(50, 50, fill=False, dx=5.0, dy=-5.0, ctx=None) ==
                         np.array(tgt_b_outline_dxy)).all())

        # Filled
        tgt_b_filled = draw_polygon_image((r @ aa_verts_xy.T).T + np.array([15.0, 25.0]),
                                          outline=1, fill=1, image_size=(50,50))
        self.assertTrue((b.render_mask(50, 50, fill=True, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_b_filled)).all())

        # Filled, offset
        tgt_b_filled_dxy = draw_polygon_image((r @ aa_verts_xy.T).T + np.array([20.0, 20.0]),
                                              outline=1, fill=1, image_size=(50,50))
        self.assertTrue((b.render_mask(50, 50, fill=True, dx=5.0, dy=-5.0, ctx=None) ==
                         np.array(tgt_b_filled_dxy)).all())

    def test_uv_points_to_params(self):
        a_cen_xy, a_rad1, a_rad2, a_ori = labelling_tool.OrientedEllipseLabel.uv_points_to_params(
            np.array([[5.0, 5.0], [15.0, 15.0]]), np.array([12.0, 8.0]))
        self.assertTrue(np.allclose(a_cen_xy, np.array([10.0, 10.0])))
        self.assertTrue(np.allclose(a_rad1, 5.0 * math.sqrt(2.0)))
        self.assertTrue(np.allclose(a_rad2, 2.0 * math.sqrt(2.0)))
        self.assertTrue(np.allclose(a_ori, np.radians(45.0)))

    def test_to_json(self):
        a = labelling_tool.OrientedEllipseLabel(
            centre_xy=np.array([15.0, 25.0]), radius1=10.0, radius2=3.0, orientation_rad=math.radians(30.0),
            object_id='abc_123', classification='cls_a', source='manual', anno_data={'purpose': 'test'})

        self.assertEqual(a.to_json(),
                         dict(label_type='oriented_ellipse', centre=dict(x=15.0, y=25.0), radius1=10.0,
                              radius2=3.0, orientation_radians=math.radians(30.0),
                              object_id='abc_123', label_class='cls_a', source='manual',
                              anno_data={'purpose': 'test'}))

    def test_from_uv_points(self):
        a = labelling_tool.OrientedEllipseLabel.new_instance_from_uv_points(
            np.array([[5.0, 5.0], [15.0, 15.0]]), np.array([12.0, 8.0]),
            object_id='abc_123', classification='cls_a', source='manual', anno_data={'purpose': 'test'})
        self.assertTrue(np.allclose(a.centre_xy, np.array([10.0, 10.0])))
        self.assertTrue(np.allclose(a.radius1, 5.0 * math.sqrt(2.0)))
        self.assertTrue(np.allclose(a.radius2, 2.0 * math.sqrt(2.0)))
        self.assertTrue(np.allclose(a.orientation_rad, np.radians(45.0)))
        self.assertEqual(a.object_id, 'abc_123')
        self.assertEqual(a.classification, 'cls_a')
        self.assertEqual(a.source, 'manual')
        self.assertEqual(a.anno_data, {'purpose': 'test'})

    def test_from_json(self):
        obj_tab = labelling_tool.ObjectTable('abc')
        js_a = dict(label_type='oriented_ellipse', centre=dict(x=15.0, y=25.0), radius1=10.0,
                    radius2=3.0, orientation_radians=math.radians(30.0),
                    object_id='abc_123', label_class='cls_a', source='manual', anno_data={'purpose': 'test'})
        a = labelling_tool.AbstractLabel.from_json(js_a, obj_tab)
        self.assertTrue(isinstance(a, labelling_tool.OrientedEllipseLabel))
        self.assertTrue((a.centre_xy == np.array([15.0, 25.0])).all())
        self.assertEqual(a.radius1, 10.0)
        self.assertEqual(a.radius2, 3.0)
        self.assertEqual(a.orientation_rad, math.radians(30.0))
        self.assertEqual(a.object_id, 'abc_123')
        self.assertEqual(a.classification, 'cls_a')
        self.assertEqual(a.source, 'manual')
        self.assertEqual(a.anno_data, {'purpose': 'test'})


class GroupLabelTestCase(TestCase):
    def test_constructor(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]),
                                    object_id='abc_123', classification='cls_a',
                                    source='manual', anno_data={'purpose': 'test'})
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        b = labelling_tool.PolygonLabel(regions=[inner_rect], object_id='abc_124', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        ab = labelling_tool.GroupLabel(component_labels=[a, b],
                                       object_id='grp_1', classification='cls_c',
                                       source='manual', anno_data={'purpose': 'all'})
        self.assertEqual(len(ab), 2)
        self.assertIs(ab[0], a)
        self.assertIs(ab[1], b)
        self.assertEqual(ab.classification, 'cls_c')
        self.assertEqual(ab.source, 'manual')
        self.assertEqual(ab.anno_data, {'purpose': 'all'})

    def test_flatten(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]),
                                    object_id='abc_123', classification='cls_a',
                                    source='manual', anno_data={'purpose': 'test'})
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        b = labelling_tool.PolygonLabel(regions=[inner_rect], object_id='abc_124', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        ab = labelling_tool.GroupLabel(component_labels=[a, b],
                                       object_id='grp_1', classification='cls_c',
                                       source='manual', anno_data={'purpose': 'all'})
        self.assertEqual(list(ab.flatten()), [a, b, ab])

    def test_bounding_box(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 30.0]), size_xy=np.array([8.0, 12.0]),
                                    object_id='abc_123', classification='cls_a',
                                    source='manual', anno_data={'purpose': 'test'})
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        b = labelling_tool.PolygonLabel(regions=[inner_rect], object_id='abc_124', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        ab = labelling_tool.GroupLabel(component_labels=[a, b],
                                       object_id='grp_1', classification='cls_c',
                                       source='manual', anno_data={'purpose': 'all'})
        self.assertTrue((a.bounding_box()[0] == np.array([11.0, 24.0])).all())
        self.assertTrue((a.bounding_box()[1] == np.array([19.0, 36.0])).all())
        self.assertTrue((b.bounding_box()[0] == np.array([20.0, 20.0])).all())
        self.assertTrue((b.bounding_box()[1] == np.array([30.0, 30.0])).all())
        self.assertTrue((ab.bounding_box()[0] == np.array([11.0, 20.0])).all())
        self.assertTrue((ab.bounding_box()[1] == np.array([30.0, 36.0])).all())

    def test_warped(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]))
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        b = labelling_tool.PolygonLabel(regions=[inner_rect])
        ab = labelling_tool.GroupLabel(component_labels=[a, b])

        ab_7 = ab.warped(lambda p_xy: p_xy + 7.0)
        self.assertTrue((ab_7[0].centre_xy == np.array([22.0, 32.0])).all())
        self.assertTrue((ab_7[0].size_xy == np.array([8.0, 12.0])).all())
        self.assertTrue((ab_7[1].regions[0] == (inner_rect + 7)).all())

        # Rotation matrix
        theta = np.radians(20.0)
        c = np.cos(theta)
        s = np.sin(theta)
        r = np.array([[c, -s],
                      [s, c]])
        ab_r = ab.warped(lambda p_xy: (r @ p_xy.T).T)
        self.assertTrue(np.allclose(ab_r[0].centre_xy, r @ np.array([15.0, 25.0])))
        self.assertTrue(np.allclose(ab_r[0].size_xy, np.array([8.0 * c + 12.0 * s, 12.0 * c + 8.0 * s])))
        self.assertTrue(np.allclose(ab_r[1].regions, (r @ inner_rect.T).T))

    def test_render_mask(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]))
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        b = labelling_tool.PolygonLabel(regions=[inner_rect])
        ab = labelling_tool.GroupLabel(component_labels=[a, b])

        # Outlined
        tgt_ab_outline = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_ab_outline).rectangle([(11.0, 19.0), (19.0, 31.0)], outline=1, fill=0)
        ImageDraw.Draw(tgt_ab_outline).polygon([tuple(v) for v in inner_rect], outline=1, fill=0)
        self.assertTrue((ab.render_mask(50, 50, fill=False, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_ab_outline)).all())

        # Outlined, offset
        tgt_ab_outline_dxy = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_ab_outline_dxy).rectangle([(16.0, 14.0), (24.0, 26.0)], outline=1, fill=0)
        ImageDraw.Draw(tgt_ab_outline_dxy).polygon([tuple(v) for v in (inner_rect + np.array([5, -5]))], outline=1, fill=0)
        self.assertTrue((ab.render_mask(50, 50, fill=False, dx=5.0, dy=-5.0, ctx=None) ==
                         np.array(tgt_ab_outline_dxy)).all())

        # Filled
        tgt_ab_filled = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_ab_filled).rectangle([(11.0, 19.0), (19.0, 31.0)], outline=1, fill=1)
        ImageDraw.Draw(tgt_ab_filled).polygon([tuple(v) for v in inner_rect], outline=1, fill=1)
        self.assertTrue((ab.render_mask(50, 50, fill=True, dx=0.0, dy=0.0, ctx=None) ==
                         np.array(tgt_ab_filled)).all())

        # Filled, offset
        tgt_ab_filled_dxy = Image.new('L', (50, 50), 0)
        ImageDraw.Draw(tgt_ab_filled_dxy).rectangle([(16.0, 14.0), (24.0, 26.0)], outline=1, fill=1)
        ImageDraw.Draw(tgt_ab_filled_dxy).polygon([tuple(v) for v in (inner_rect + np.array([5, -5]))],
                                                  outline=1, fill=1)
        self.assertTrue((ab.render_mask(50, 50, fill=True, dx=5.0, dy=-5.0, ctx=None) ==
                         np.array(tgt_ab_filled_dxy)).all())

    def test_to_json(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]),
                                    object_id='abc_123', classification='cls_a',
                                    source='manual', anno_data={'purpose': 'test'})
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        inner_js = [dict(x=p[0], y=p[1]) for p in inner_rect]
        b = labelling_tool.PolygonLabel(regions=[inner_rect], object_id='abc_124', classification='cls_b',
                                        source='manual2', anno_data={'purpose': 'test2'})
        ab = labelling_tool.GroupLabel(component_labels=[a, b],
                                       object_id='grp_1', classification='cls_c',
                                       source='manual3', anno_data={'purpose': 'test3'})

        js_a = dict(label_type='box', centre=dict(x=15.0, y=25.0), size=dict(x=8.0, y=12.0),
                    object_id='abc_123', label_class='cls_a', source='manual', anno_data={'purpose': 'test'})
        js_b = dict(label_type='polygon', regions=[inner_js], object_id='abc_124',
                    label_class='cls_b', source='manual2', anno_data={'purpose': 'test2'})
        self.assertEqual(ab.to_json()['component_models'][0], js_a)
        self.assertEqual(ab.to_json()['component_models'][1], js_b)
        self.assertEqual(ab.to_json(),
                         dict(label_type='group', component_models=[js_a, js_b], object_id='grp_1',
                              label_class='cls_c', source='manual3', anno_data={'purpose': 'test3'}))

    def test_from_json(self):
        obj_tab = labelling_tool.ObjectTable('abc')
        js_a = dict(label_type='box', centre=dict(x=15.0, y=25.0), size=dict(x=8.0, y=12.0),
                    object_id='abc_123', label_class='cls_a', source='manual', anno_data={'purpose': 'test'})

        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        inner_js = [dict(x=p[0], y=p[1]) for p in inner_rect]
        js_b = dict(label_type='polygon', regions=[inner_js], object_id='abc_124',
                    label_class='cls_b', source='manual2', anno_data={'purpose': 'test2'})

        js_ab = dict(label_type='group', component_models=[js_a, js_b], object_id='grp_1',
                     label_class='cls_c', source='manual3', anno_data={'purpose': 'test3'})
        ab = labelling_tool.AbstractLabel.from_json(js_ab, obj_tab)

        self.assertTrue(isinstance(ab, labelling_tool.GroupLabel))
        self.assertTrue((ab[0].centre_xy == np.array([15.0, 25.0])).all())
        self.assertTrue((ab[0].size_xy == np.array([8.0, 12.0])).all())
        self.assertEqual(ab[0].object_id, 'abc_123')
        self.assertEqual(ab[0].classification, 'cls_a')
        self.assertEqual(ab[0].source, 'manual')
        self.assertEqual(ab[0].anno_data, {'purpose': 'test'})
        self.assertEqual(len(ab[1].regions), 1)
        self.assertTrue((ab[1].regions[0] == inner_rect).all())
        self.assertEqual(ab[1].object_id, 'abc_124')
        self.assertEqual(ab[1].classification, 'cls_b')
        self.assertEqual(ab[1].source, 'manual2')
        self.assertEqual(ab[1].anno_data, {'purpose': 'test2'})
        self.assertEqual(len(ab), 2)
        self.assertEqual(ab.object_id, 'grp_1')
        self.assertEqual(ab.classification, 'cls_c')
        self.assertEqual(ab.source, 'manual3')
        self.assertEqual(ab.anno_data, {'purpose': 'test3'})

    def test_flatten_json(self):
        js_a = dict(label_type='box', centre=dict(x=15.0, y=25.0), size=dict(x=8.0, y=12.0),
                    object_id='abc_123', label_class='cls_a', source='manual', anno_data={'purpose': 'test'})
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        inner_js = [dict(x=p[0], y=p[1]) for p in inner_rect]
        js_b = dict(label_type='polygon', regions=[inner_js], object_id='abc_124',
                    label_class='cls_b', source='manual2', anno_data={'purpose': 'test2'})
        js_ab = dict(label_type='group', component_models=[js_a, js_b], object_id='grp_1',
                     label_class='cls_c', source='manual3', anno_data={'purpose': 'test3'})
        labs_js = list(labelling_tool.AbstractLabel.flatten_json(js_ab))
        self.assertEqual(labs_js, [js_a, js_b, js_ab])


class ImageLabelsTestCase(TestCase):
    def test_constructor(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]),
                                    object_id='abc_123', classification='cls_a',
                                    source='manual', anno_data={'purpose': 'test'})
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        b = labelling_tool.PolygonLabel(regions=[inner_rect], object_id='abc_124', classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        ab = labelling_tool.GroupLabel(component_labels=[a, b],
                                       object_id='grp_1', classification='cls_c',
                                       source='manual', anno_data={'purpose': 'all'})
        labels = labelling_tool.ImageLabels([ab])
        self.assertEqual(len(labels), 1)
        self.assertIs(labels[0], ab)
        self.assertIs(labels[0][0], a)
        self.assertIs(labels[0][1], b)
        self.assertIs(labels['abc_123'], a)
        self.assertIs(labels['abc_124'], b)
        self.assertIs(labels['grp_1'], ab)

    def test_constructor_b(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]),
                                    classification='cls_a',
                                    source='manual', anno_data={'purpose': 'test'})
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        b = labelling_tool.PolygonLabel(regions=[inner_rect], classification='cls_a',
                                        source='manual', anno_data={'purpose': 'test'})
        ab = labelling_tool.GroupLabel(component_labels=[a, b],
                                       classification='cls_c',
                                       source='manual', anno_data={'purpose': 'all'})
        labels = labelling_tool.ImageLabels([ab], id_prefix='abc')
        self.assertEqual(a.object_id, 'abc__1')
        self.assertEqual(b.object_id, 'abc__2')
        self.assertEqual(ab.object_id, 'abc__3')
        self.assertEqual(len(labels), 1)
        self.assertIs(labels[0], ab)
        self.assertIs(labels[0][0], a)
        self.assertIs(labels[0][1], b)
        self.assertIs(labels['abc__1'], a)
        self.assertIs(labels['abc__2'], b)
        self.assertIs(labels['abc__3'], ab)

    def test_flatten(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]),
                                    classification='cls_a')
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        b = labelling_tool.PolygonLabel(regions=[inner_rect], classification='cls_a')
        ab = labelling_tool.GroupLabel(component_labels=[a, b], classification='cls_c')
        labels = labelling_tool.ImageLabels([ab], id_prefix='abc')
        self.assertEqual(list(labels.flatten()), [a, b, ab])

    def test_label_class_histogram(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]),
                                    classification='cls_a')
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        b = labelling_tool.PolygonLabel(regions=[inner_rect], classification='cls_b')
        outer_rect = np.array([[10.0, 10.0], [40.0, 10.0], [40.0, 40.0], [10.0, 40.0]])
        c = labelling_tool.PolygonLabel(regions=[outer_rect, inner_rect], classification='cls_a')
        ab = labelling_tool.GroupLabel(component_labels=[a, b], classification='cls_c')
        self.assertEqual(labelling_tool.ImageLabels([ab]).label_class_histogram(),
                         {'cls_c': 1})
        self.assertEqual(labelling_tool.ImageLabels([a, b, c]).label_class_histogram(),
                         {'cls_a': 2, 'cls_b': 1})

    def test_replace_label_classes(self):
        a = labelling_tool.BoxLabel(centre_xy=np.array([15.0, 25.0]), size_xy=np.array([8.0, 12.0]),
                                    classification='cls_a')
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        b = labelling_tool.PolygonLabel(regions=[inner_rect], classification='cls_b')
        outer_rect = np.array([[10.0, 10.0], [40.0, 10.0], [40.0, 40.0], [10.0, 40.0]])
        c = labelling_tool.PolygonLabel(regions=[outer_rect, inner_rect], classification='cls_a')
        ab = labelling_tool.GroupLabel(component_labels=[a, b], classification='cls_c')
        labels = labelling_tool.ImageLabels([ab, c])
        self.assertEqual(a.classification, 'cls_a')
        self.assertEqual(b.classification, 'cls_b')
        self.assertEqual(c.classification, 'cls_a')
        self.assertEqual(ab.classification, 'cls_c')
        labels.replace_label_classes({'cls_a': 'new_a', 'cls_c': 'new_c'})
        self.assertEqual(a.classification, 'new_a')
        self.assertEqual(b.classification, 'cls_b')
        self.assertEqual(c.classification, 'new_a')
        self.assertEqual(ab.classification, 'new_c')

    def test_replace_label_classes_json(self):
        inner_rect = np.array([[20.0, 20.0], [30.0, 20.0], [30.0, 30.0], [20.0, 30.0]])
        outer_rect = np.array([[10.0, 10.0], [40.0, 10.0], [40.0, 40.0], [10.0, 40.0]])
        inner_js = [dict(x=p[0], y=p[1]) for p in inner_rect]
        outer_js = [dict(x=p[0], y=p[1]) for p in outer_rect]
        js_a = dict(label_type='box', centre=dict(x=15.0, y=25.0), size=dict(x=8.0, y=12.0),
                    object_id='abc_123', label_class='cls_a')
        js_b = dict(label_type='polygon', regions=[inner_js], object_id='abc_124',
                    label_class='cls_b')
        js_c = dict(label_type='polygon', regions=[outer_js, inner_js], object_id='abc_125',
                    label_class='cls_a')
        js_ab = dict(label_type='group', component_models=[js_a, js_b], object_id='grp_1',
                     label_class='cls_c')
        js_labels = [js_ab, js_c]

        self.assertEqual(js_labels[0]['component_models'][0]['label_class'], 'cls_a')
        self.assertEqual(js_labels[0]['component_models'][1]['label_class'], 'cls_b')
        self.assertEqual(js_labels[1]['label_class'], 'cls_a')
        self.assertEqual(js_labels[0]['label_class'], 'cls_c')
        labelling_tool.ImageLabels.replace_label_classes_json(js_labels, {'cls_a': 'new_a', 'cls_c': 'new_c'})
        self.assertEqual(js_labels[0]['component_models'][0]['label_class'], 'new_a')
        self.assertEqual(js_labels[0]['component_models'][1]['label_class'], 'cls_b')
        self.assertEqual(js_labels[1]['label_class'], 'new_a')
        self.assertEqual(js_labels[0]['label_class'], 'new_c')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-09-11 08:54
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('example_labeller', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='imagewithlabels',
            name='image',
            field=models.ImageField(blank=True, upload_to=''),
        ),
        migrations.AlterField(
            model_name='imagewithlabels',
            name='labels',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='image', to='image_labelling_tool.Labels'),
        ),
    ]



import os
import mimetypes
import json
import datetime
from django.core.management.base import BaseCommand, CommandError
from django.core.files import File
from django.db import transaction
from image_labelling_tool import labelling_tool
from image_labelling_tool import models as lt_models
from ... import models

class Command(BaseCommand):
    help = 'Populates the image database from a directory'

    def add_arguments(self, parser):
        parser.add_argument('dir', type=str)

    def handle(self, *args, **options):
        images_dir = options['dir']
        image_and_label_files = []
        for filename in os.listdir(images_dir):
            path = os.path.join(images_dir, filename)
            if os.path.isfile(path):
                mt, encoding = mimetypes.guess_type(path)
                if mt is not None and mt.startswith('image/'):
                    image_path = path
                    labels_path = os.path.splitext(path)[0] + '__labels.json'
                    if os.path.exists(labels_path) and os.path.isfile(labels_path):
                        image_and_label_files.append((image_path, labels_path))
                    else:
                        image_and_label_files.append((image_path, None))

        with transaction.atomic():
            for image_path, labels_path in image_and_label_files:
                if labels_path is not None:
                    self.stdout.write('Adding image {} with labels from {}'.format(image_path, labels_path))
                    wrapped_labels = labelling_tool.WrappedImageLabels.from_file(labels_path)
                    completed_tasks = wrapped_labels.completed_tasks

                    # Convert task names to instances
                    tasks = list(lt_models.LabellingTask.objects.filter(name__in=completed_tasks).distinct())

                    labels_model = lt_models.Labels(
                        labels_json_str=json.dumps(wrapped_labels.labels_json),
                        creation_date=datetime.date.today())
                    labels_model.save()
                    if len(tasks) > 0:
                        labels_model.completed_tasks.set(tasks)
                        labels_model.save()
                else:
                    self.stdout.write('Adding image {}'.format(image_path))
                    labels_model = lt_models.Labels(creation_date=datetime.date.today())
                    labels_model.save()

                image_model = models.ImageWithLabels(labels=labels_model)
                image_model.image.save(os.path.basename(image_path),
                                       File(open(image_path, 'rb')))
                image_model.save()



import os, datetime, json, tempfile, zipfile
import celery.result

from PIL import Image

from dateutil.tz import tzlocal

from django.shortcuts import render, redirect, get_object_or_404
from django.views.decorators.csrf import ensure_csrf_cookie
from django.db import transaction
from django.db.models import Q
from django.core.files import File

from django.conf import settings
import django.utils.timezone

from image_labelling_tool import labelling_tool
from image_labelling_tool import models as lt_models
from image_labelling_tool import labelling_tool_views, schema_editor_views

from . import models, tasks, forms


@ensure_csrf_cookie
def home(request):
    upload_form = forms.ImageUploadForm()

    if 'example_labeller_message' in request.session:
        message = request.session.pop('example_labeller_message')
    else:
        message = None

    context = {
        'upload_form': upload_form,
        'message': message,
        'num_images': len(models.ImageWithLabels.objects.all())
    }
    return render(request, 'index.html', context)


@ensure_csrf_cookie
def upload_images(request):
    if request.method == 'POST':
        upload_form = forms.ImageUploadForm(request.POST, request.FILES)

        if upload_form.is_valid():
            uploaded_file = upload_form.cleaned_data['file']

            print(uploaded_file.content_type)

            if uploaded_file.content_type in {'image/jpeg', 'image/png'}:
                # Single image upload

                # Blank labels
                labels_model = lt_models.Labels(creation_date=datetime.date.today())
                labels_model.save()

                image_model = models.ImageWithLabels(labels=labels_model)
                image_model.image.save(os.path.basename(uploaded_file.name), uploaded_file)
                image_model.save()
            elif uploaded_file.content_type in {'application/zip', 'application/x-zip-compressed'}:
                # ZIP file

                # Write to a temporary file
                handle, upload_path =tempfile.mkstemp()
                os.close(handle)
                os.remove(upload_path)

                with open(upload_path, 'wb+') as f_dest:
                    for chunk in uploaded_file.chunks():
                        f_dest.write(chunk)

                # Load the ZIP and get its contents
                z = zipfile.ZipFile(upload_path, 'r')

                # Pair image files with corresponding label files
                name_to_image_and_labels = {}
                for filename_and_ext in z.namelist():
                    filename, ext = os.path.splitext(filename_and_ext)
                    if ext.lower() in {'.png', '.jpg', '.jpeg'}:
                        entry = name_to_image_and_labels.setdefault(filename, dict(image=None, labels=None))
                        entry['image'] = filename_and_ext
                    elif ext.lower() == '.json':
                        if filename.endswith('__labels'):
                            filename = filename[:-8]
                        entry = name_to_image_and_labels.setdefault(filename, dict(image=None, labels=None))
                        entry['labels'] = filename_and_ext

                # Add all images using a single transaction
                with transaction.atomic():
                    for name, entry in name_to_image_and_labels.items():
                        # Entry is only valid if there is an image file
                        if entry['image'] is not None:
                            valid_image = False
                            # Attempt to open the image to ensure its valid
                            with z.open(entry['image'], mode='r') as f_img:
                                try:
                                    im = Image.open(f_img)
                                except IOError:
                                    pass
                                else:
                                    valid_image = True
                                    im.close()

                            if valid_image:
                                labels_model = None
                                # See if we have a labels file
                                if entry['labels'] is not None:
                                    # Open the labels
                                    with z.open(entry['labels'], mode='r') as f_labels:
                                        try:
                                            wrapped_labels = json.load(f_labels)
                                        except IOError:
                                            pass
                                        else:
                                            # Get the modification date and time of the labels file
                                            z_info = z.getinfo(entry['labels'])
                                            year, month, day, hour, minute, second = z_info.date_time
                                            creation_date = datetime.date(
                                                year=year, month=month, day=day)
                                            modification_datetime = datetime.datetime(
                                                year=year, month=month, day=day, hour=hour, minute=minute,
                                                second=second, tzinfo=tzlocal())
                                            if request.user.is_authenticated:
                                                modification_user = request.user
                                            else:
                                                modification_user = None

                                            # Unwrap the labels
                                            labels, complete = labelling_tool.PersistentLabelledImage._unwrap_labels(
                                                wrapped_labels)
                                            complete = complete if isinstance(complete, bool) else False

                                            # Build labels model
                                            labels_model = lt_models.Labels(
                                                labels_json_str=json.dumps(labels), complete=complete,
                                                creation_date=creation_date,
                                                last_modified_datetime=modification_datetime,
                                                last_modified_by=modification_user)
                                            labels_model.save()

                                if labels_model is None:
                                    # No labels loaded; create an empty labels model
                                    labels_model = lt_models.Labels(creation_date=datetime.date.today())
                                    labels_model.save()

                                image_model = models.ImageWithLabels(labels=labels_model)
                                image_model.image.save(os.path.basename(entry['image']),
                                                       File(z.open(entry['image'], mode='r')))
                                image_model.save()
            else:
                # Unknown type; put message in session
                request.session['example_labeller_message'] = 'unknown_upload_filetype'
    return redirect('example_labeller:home')




@ensure_csrf_cookie
def tool(request):
    image_descriptors = [labelling_tool.image_descriptor(
            image_id=img.id, url=img.image.url,
            width=img.image.width, height=img.image.height) for img in models.ImageWithLabels.objects.all()]

    try:
        schema = lt_models.LabellingSchema.objects.get(name='default')
    except lt_models.LabellingSchema.DoesNotExist:
        schema_js = dict(colour_schemes=[], label_class_groups=[])
    else:
        schema_js = schema.json_for_tool()

    context = {
        'labelling_schema': schema_js,
        'image_descriptors': image_descriptors,
        'initial_image_index': str(0),
        'labelling_tool_config': settings.LABELLING_TOOL_CONFIG,
        'tasks': lt_models.LabellingTask.objects.filter(enabled=True).order_by('order_key'),
        'anno_controls': [c.to_json() for c in settings.ANNO_CONTROLS],
        'enable_locking': settings.LABELLING_TOOL_ENABLE_LOCKING,
        'dextr_available': settings.LABELLING_TOOL_DEXTR_AVAILABLE,
        'dextr_polling_interval': settings.LABELLING_TOOL_DEXTR_POLLING_INTERVAL,
    }
    return render(request, 'tool.html', context)


class LabellingToolAPI (labelling_tool_views.LabellingToolViewWithLocking):
    def get_labels(self, request, image_id_str, *args, **kwargs):
        image = get_object_or_404(models.ImageWithLabels, id=int(image_id_str))
        return image.labels

    def get_unlocked_image_id(self, request, image_ids, *args, **kwargs):
        unlocked_labels = lt_models.Labels.objects.unlocked()
        unlocked_q = Q(id__in=image_ids, labels__in=unlocked_labels)
        # TODO FOR YOUR APPLICATION
        # filter images for those accessible to the user to guard against maliciously crafted requests
        accessible_q = Q()
        unlocked_imgs = models.ImageWithLabels.objects.filter(unlocked_q & accessible_q).distinct()
        first_unlocked = unlocked_imgs.first()
        return first_unlocked.id if first_unlocked is not None else None

    def dextr_request(self, request, image_id_str, dextr_id, dextr_points):
        """
        :param request: HTTP request
        :param image_id_str: image ID that identifies the image that we are labelling
        :param dextr_id: an ID number the identifies the DEXTR request
        :param dextr_points: the 4 points as a list of 2D vectors ({'x': <x>, 'y': <y>}) in the order
            top edge, left edge, bottom edge, right edge
        :return: contours/regions a list of lists of 2D vectors, each of which is {'x': <x>, 'y': <y>}
        """
        if settings.LABELLING_TOOL_DEXTR_AVAILABLE:
            image = get_object_or_404(models.ImageWithLabels, id=int(image_id_str))
            cel_result = tasks.dextr.delay(image.image.path, dextr_points)
            dtask = models.DextrTask(image=image, image_id_str=image_id_str, dextr_id=dextr_id, celery_task_id=cel_result.id)
            dtask.save()
        return None

    def dextr_poll(self, request, image_id_str, dextr_ids):
        """
        :param request: HTTP request
        :param image_id_str: image ID that identifies the image that we are labelling
        :param dextr_ids: The DEXTR request IDs that the client is interested in
        :return: a list of dicts where each dict takes the form:
            {
                'image_id': image ID string that identifies the image that the label applies to
                'dextr_id': the ID number that identifies the dextr job/request
                'regions': contours/regions a list of lists of 2D vectors, each of which is {'x': <x>, 'y': <y>}
            }
        """
        to_remove = []
        dextr_labels = []
        for dtask in models.DextrTask.objects.filter(image__id=image_id_str, dextr_id__in=dextr_ids):
            uuid = dtask.celery_task_id
            res = celery.result.AsyncResult(uuid)
            if res.ready():
                try:
                    regions = res.get()
                except:
                    # An error occurred during the DEXTR task; nothing we can do
                    pass
                else:
                    dextr_label = dict(image_id=dtask.image_id_str, dextr_id=dtask.dextr_id, regions=regions)
                    dextr_labels.append(dextr_label)
                to_remove.append(dtask)

        # Remove old tasks
        oldest = django.utils.timezone.now() - datetime.timedelta(minutes=10)
        for old_task in models.DextrTask.objects.filter(creation_timestamp__lt=oldest):
            to_remove.append(old_task)

        for r in to_remove:
            r.delete()

        return dextr_labels


@ensure_csrf_cookie
def schema_editor(request):
    context = {'schema': lt_models.LabellingSchema.objects.get(name='default')}
    return render(request, 'schema_editor.html', context)


class SchemaEditorAPI (schema_editor_views.SchemaEditorView):
    def get_schema(self, request, *args, **kwargs):
        return lt_models.LabellingSchema.objects.get(name='default')


"""
Django settings for example_labeller_app project.

Generated by 'django-admin startproject' using Django 1.8.2.

For more information on this file, see
https://docs.djangoproject.com/en/1.8/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.8/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os
from image_labelling_tool import labelling_tool

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.8/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '3+p2$qln6o1ws1c)6o!+o+p%ql1n!+tt@wp)g5!pfgliqld)yo'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'image_labelling_tool.apps.ImageLabellingToolConfig',
    'example_labeller.apps.ExampleLabellerConfig',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'example_labeller_app.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.static',
            ],
        },
    },
]

WSGI_APPLICATION = 'example_labeller_app.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.8/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}


# Internationalization
# https://docs.djangoproject.com/en/1.8/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.8/howto/static-files/

STATIC_URL = '/static/'

MEDIA_ROOT = os.path.join(BASE_DIR, 'media')
MEDIA_URL = '/media/'

CSRF_COOKIE_SECURE = False


#
#
# Colour schemes and label classes are stored in the database
#
#


# Annotation controls
# Labels may also have optional meta-data associated with them
# You could use this for e.g. indicating if an object is fully visible, mostly visible or significantly obscured.
# You could also indicate quality (e.g. blurriness, etc)
# There are four types of annotation. They have some common properties:
#   - name: symbolic name (Python identifier)
#   - label_text: label text in UI
#   Check boxes, radio buttons and popup menus also have:
#     - visibility_label_text: [optional] if provided, label visibility can be filtered by this annotation value,
#       in which case a drop down will appear in the UI allowing the user to select a filter value
#       that will hide/show labels accordingly
# Control types:
# Check box (boolean value):
#   `labelling_tool.AnnoControlCheckbox`; only the 3 common parameters listed above
# Radio button (choice from a list):
#   `labelling_tool.AnnoControlRadioButtons`; the 3 common parameters listed above and:
#       choices: list of `labelling_tool.AnnoControlRadioButtons.choice` that provide:
#           value: symbolic value name for choice
#           tooltip: extra information for user
#       label_on_own_line [optional]: if True, place the label and the buttons on a separate line in the UI
# Popup menu (choice from a grouped list):
#   `labelling_tool.AnnoControlPopupMenu`; the 3 common parameters listed above and:
#       groups: list of groups `labelling_tool.AnnoControlPopupMenu.group`:
#           label_text: label text in UI
#           choices: list of `labelling_tool.AnnoControlPopupMenu.choice` that provide:
#               value: symbolic value name for choice
#               label_text: choice label text in UI
#               tooltip: extra information for user
# Text (free form plain text):
#   `labelling_tool.AnnoControlText`; only the 2 common parameters listed above and:
#       - multiline: boolean; if True a text area will be used, if False a single line text entry
ANNO_CONTROLS = [
    labelling_tool.AnnoControlCheckbox('good_quality', 'Good quality',
                                       visibility_label_text='Filter by good quality'),
    labelling_tool.AnnoControlRadioButtons('visibility', 'Visible', choices=[
        labelling_tool.AnnoControlRadioButtons.choice(value='full', label_text='Fully',
                                                      tooltip='Object is fully visible'),
        labelling_tool.AnnoControlRadioButtons.choice(value='mostly', label_text='Mostly',
                                                      tooltip='Object is mostly visible'),
        labelling_tool.AnnoControlRadioButtons.choice(value='obscured', label_text='Obscured',
                                                      tooltip='Object is significantly obscured'),
    ], label_on_own_line=False, visibility_label_text='Filter by visibility'),
    labelling_tool.AnnoControlPopupMenu('material', 'Material', groups=[
        labelling_tool.AnnoControlPopupMenu.group(label_text='Artifical/buildings', choices=[
            labelling_tool.AnnoControlPopupMenu.choice(value='concrete', label_text='Concrete',
                                                       tooltip='Concrete objects'),
            labelling_tool.AnnoControlPopupMenu.choice(value='plastic', label_text='Plastic',
                                                       tooltip='Plastic objects'),
            labelling_tool.AnnoControlPopupMenu.choice(value='asphalt', label_text='Asphalt',
                                                       tooltip='Road, pavement, etc.'),
        ]),
        labelling_tool.AnnoControlPopupMenu.group(label_text='Flat natural', choices=[
            labelling_tool.AnnoControlPopupMenu.choice(value='grass', label_text='Grass',
                                                       tooltip='Grass covered ground'),
            labelling_tool.AnnoControlPopupMenu.choice(value='water', label_text='Water', tooltip='Water/lake')]),
        labelling_tool.AnnoControlPopupMenu.group(label_text='Vegetation', choices=[
            labelling_tool.AnnoControlPopupMenu.choice(value='trees', label_text='Trees', tooltip='Trees'),
            labelling_tool.AnnoControlPopupMenu.choice(value='shrubbery', label_text='Shrubs',
                                                       tooltip='Shrubs/bushes'),
            labelling_tool.AnnoControlPopupMenu.choice(value='flowers', label_text='Flowers',
                                                       tooltip='Flowers'),
            labelling_tool.AnnoControlPopupMenu.choice(value='ivy', label_text='Ivy', tooltip='Ivy')]),
    ], visibility_label_text='Filter by material'),
    # labelling_tool.AnnoControlText('comment', 'Comment', multiline=False),
]

# Configuration
LABELLING_TOOL_CONFIG = {
    'useClassSelectorPopup': True,
    'tools': {
        'imageSelector': True,
        'labelClassSelector': True,
        'brushSelect': True,
        'labelClassFilter': True,
        'drawPointLabel': False,
        'drawBoxLabel': True,
        'drawOrientedEllipseLabel': True,
        'drawPolyLabel': True,
        'deleteLabel': True,
        'deleteConfig': {
            'typePermissions': {
                'point': True,
                'box': True,
                'polygon': True,
                'composite': True,
                'group': True,
            }
        }
    },
    'settings': {
        'brushWheelRate': 0.025,  # Change rate for brush radius (mouse wheel)
        'brushKeyRate': 2.0,    # Change rate for brush radius (keyboard)
    }
}

LABELLING_TOOL_ENABLE_LOCKING = False
LABELLING_TOOL_DEXTR_AVAILABLE = False
LABELLING_TOOL_DEXTR_POLLING_INTERVAL = 1000
LABELLING_TOOL_DEXTR_WEIGHTS_PATH = None



CELERY_BROKER_URL = 'amqp://guest@localhost//'
CELERY_RESULT_BACKEND = 'rpc://'

CELERY_ACCEPT_CONTENT = ['json']
CELERY_TASK_SERIALIZER = 'json'
CELERY_RESULT_SERIALIZER = 'json'



from django.urls import include, path

from . import views

app_name = 'example_labeller'

urlpatterns = [
    path('', views.home, name='home'),
    path('upload_images', views.upload_images, name='upload_images'),
    path('tool', views.tool, name='tool'),
    path('labelling_tool_api', views.LabellingToolAPI.as_view(), name='labelling_tool_api'),
    path('schema_editor', views.schema_editor, name='schema_editor'),
    path('schema_editor_api', views.SchemaEditorAPI.as_view(), name='schema_editor_api'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-30 09:03
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('image_labelling_tool', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ImageWithLabels',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image', models.ImageField(blank=True, upload_to=b'')),
                ('labels', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='image', to='image_labelling_tool.Labels')),
            ],
        ),
    ]
